# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import ctypes
import inspect
import os
import zlib
import json
import re
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor

import numpy as np
import torch
from torch import distributed as dist
from torch.distributed.distributed_c10d import _get_default_group

from msprobe.core.common.file_utils import FileOpen, load_json
from msprobe.core.common.const import Const
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.core.common.exceptions import MsprobeException
from msprobe.core.common.log import logger
from msprobe.core.common.utils import convert_tuple, is_int
from msprobe.core.dump.data_dump.data_processor.base import (
    BaseDataProcessor,
    ModuleBackwardInputsOutputs,
    ModuleForwardInputsOutputs,
    TensorStatInfo
)
from msprobe.pytorch.common.utils import (
    Const as PtConst,
    save_pt,
    is_recomputation,
    is_hifloat8_tensor,
    is_float8_tensor
)


is_gpu = False
try:
    import torch_npu
except ImportError:
    is_gpu = True


class TensorHandler:
    def __init__(self):
        self.has_dtensor = hasattr(dist, "tensor") and hasattr(dist.tensor, "DTensor")
        self.has_fake_tensor = hasattr(torch, "_subclasses") and hasattr(torch._subclasses, "fake_tensor")
        self.has_async_collective_tensor = hasattr(dist, "_functional_collectives") and \
                                           hasattr(dist._functional_collectives, "AsyncCollectiveTensor")
        self.has_nested_tensor = hasattr(torch, "nested") and hasattr(torch.nested, "_internal") and \
            hasattr(torch.nested._internal, "nested_tensor") and \
            hasattr(torch.nested._internal.nested_tensor, "NestedTensor")

    @staticmethod
    def free_tensor(tensor, tensor_name):
        try:
            tensor.untyped_storage().resize_(0)
        except Exception as e:
            logger.warning(f"Failed to free tensor: {tensor_name}, the detail info: {e}.")

    @staticmethod
    def get_tensor_dtype(tensor):
        if is_hifloat8_tensor(tensor):
            return PtConst.HIFLOAT8_TYPE
        return str(tensor.dtype)

    def is_dtensor(self, tensor):
        return self.has_dtensor and isinstance(tensor, dist.tensor.DTensor)

    def is_fake_tensor(self, tensor):
        return self.has_fake_tensor and isinstance(tensor, torch._subclasses.fake_tensor.FakeTensor)

    def is_async_collective_tensor(self, tensor):
        return self.has_async_collective_tensor and \
            isinstance(tensor, dist._functional_collectives.AsyncCollectiveTensor)
    
    def is_nested_tensor(self, tensor):
        return self.has_nested_tensor and \
            isinstance(tensor, torch.nested._internal.nested_tensor.NestedTensor)

    def is_empty_data(self, tensor):
        return tensor.is_meta or self.is_fake_tensor(tensor) or self.is_async_collective_tensor(tensor)

    def convert_common_tensor(self, tensor):
        if self.is_dtensor(tensor):
            return tensor.to_local()
        if self.is_fake_tensor(tensor):
            logger.debug("FakeTensor cannot be converted to torch.Tensor type.")
            return tensor
        if self.is_nested_tensor(tensor):
            logger.debug(f"For NestedTensor, collecting information from the tensor returned by .values().")
            return tensor.values()
        if is_float8_tensor(tensor):
            logger.debug(
                f"The fp8/hifp8 tensor analyzing/saving is unsupported in dump function."
                f"Casting to float for processing."
            )
            tensor = tensor.detach().float()
        return tensor

    def get_tensor_type(self, tensor):
        if self.is_dtensor(tensor):
            return Const.DTENSOR_TYPE
        if self.is_fake_tensor(tensor):
            return Const.FAKE_TENSOR_TYPE
        if self.is_async_collective_tensor(tensor):
            return Const.AC_TENSOR_TYPE
        if self.is_nested_tensor(tensor):
            return Const.NESTED_TENSOR_TYPE
        return Const.TENSOR_TYPE

    def get_dtensor_info(self, tensor):
        dtensor_info = {}
        if not self.is_dtensor(tensor):
            return dtensor_info
        if hasattr(tensor, "device_mesh") and tensor.device_mesh:
            dtensor_info.update({"device_mesh": tensor.device_mesh.mesh.tolist()})

        placements = []
        if hasattr(tensor, "placements") and isinstance(tensor.placements, Iterable):
            for placement in tensor.placements:
                if placement.is_shard() and is_int(placement.dim):
                    placements.append({"Shard": {"dim": placement.dim}})
                    continue
                if placement.is_replicate():
                    placements.append({"Replicate": {}})
                    continue
                if placement.is_partial() and isinstance(placement.reduce_op, str):
                    placements.append({"Partial": {"reduce_op": placement.reduce_op}})
        dtensor_info.update({"placements": placements})
        return dtensor_info

    def save_tensor(self, tensor, file_path):
        common_tensor = self.convert_common_tensor(tensor)
        if self.is_empty_data(common_tensor):
            logger.debug(f"Saving fake tensor or meta tensor is not supported, the current tensor is {file_path}.")
            return
        if common_tensor.untyped_storage().data_ptr() == 0:
            logger.debug(f"Saving null-pointer tensor is not supported, the current tensor is {file_path}.")
            return
        saved_tensor = common_tensor.clone().contiguous().detach()
        save_pt(saved_tensor, file_path)
        self.free_tensor(saved_tensor, file_path)


class PytorchDataProcessor(BaseDataProcessor):
    pytorch_special_type = (
        torch.device,
        torch.dtype,
        torch.Size,
        torch.Tensor,
        torch.memory_format,
        dist.ProcessGroup,
        dist.P2POp,
        dist.ReduceOp
    )
    memory_format = {
        torch.contiguous_format: "contiguous_format",
        torch.channels_last: "channels_last",
        torch.channels_last_3d: "channels_last_3d",
        torch.preserve_format: "preserve_format"
    }

    def __init__(self, config, data_writer):
        super().__init__(config, data_writer)
        self.torch_object_key = {
            "device": self.analyze_device_in_kwargs,
            "dtype": self.analyze_dtype_in_kwargs
        }
        self._async_dump_cache = {}
        self.tensor_handler = TensorHandler()
        self._crc_executor = ThreadPoolExecutor(max_workers=os.cpu_count() // 2)

    @staticmethod
    def get_md5_for_tensor(x):
        if x.dtype == torch.bfloat16:
            x = x.float()
        tensor_bytes = x.cpu().detach().numpy().tobytes()
        crc32_hash = zlib.crc32(tensor_bytes)
        return f"{crc32_hash:08x}"

    @staticmethod
    def tensor_bytes_view_cpu(t: torch.Tensor):
        """
        返回 t 在当前 dtype 下的原始字节视图（优先零拷贝）。
        需保证：t 已在 CPU 且是 contiguous。
        可能返回 memoryview 或 bytes（兜底拷贝）或者 转为numpy，均可被 zlib.crc32 接受。
        """

        nbytes = t.numel() * t.element_size()
        byte_offset = t.storage_offset() * t.element_size()

        if nbytes == 0:
            return memoryview(b"")

        storage = t.untyped_storage()

        # ctypes 指针构造 memoryview（零拷贝 FFI）
        try:
            addr = storage.data_ptr() + byte_offset
            buf = (ctypes.c_ubyte * nbytes).from_address(addr)
            mv3 = memoryview(buf)

            return mv3
        except Exception as e1:
            logger.warning(f"path_A_failed: {e1}.")

        try:
            data = ctypes.string_at(storage.data_ptr() + byte_offset, nbytes)

            return data  # bytes 也可直接用于 zlib.crc32
        except Exception as e2:
            logger.warning(f"path_B_failed: {e2}.")

        try:
            if t.dtype == torch.bfloat16:
                t = t.float()
            data = t.numpy()

            return data
        except Exception as e3:
            logger.warning(f"path_C_failed: {e3}.")
            return memoryview(b"")

    @staticmethod
    def compute_crc32_from_tensor(t: torch.Tensor) -> str:
        """
        直接对 Tensor 原始字节做 CRC32。
        :
        - "raw": 保持 bfloat16 原始 16bit 字节（推荐，避免升精/增容）
        """

        # 取得字节视图（含多级回退），然后做 CRC
        mv = PytorchDataProcessor.tensor_bytes_view_cpu(t)

        crc = zlib.crc32(mv)

        return f"{crc:08x}"

    @staticmethod
    def analyze_device_in_kwargs(element):
        single_arg = {}
        single_arg.update({'type': "torch.device"})
        if isinstance(element, (int, str)):
            single_arg.update({"value": element})
        elif isinstance(element, torch.device):
            if hasattr(element, "index"):
                device_value = element.type + ":" + str(element.index)
            else:
                device_value = element.type
            single_arg.update({"value": device_value})
        else:
            logger.debug(f"Device type {type(element)} is not supported.")
        return single_arg

    @staticmethod
    def analyze_dtype_in_kwargs(element):
        return {"type": "torch.dtype", "value": str(element)}

    @staticmethod
    def process_group_hash(arg):
        group_ranks = dist.get_process_group_ranks(arg)
        group_ranks_hash = zlib.crc32(str(group_ranks).encode('utf-8'))
        return f"{group_ranks_hash:08x}"

    @staticmethod
    def is_hookable_element(element):
        return (hasattr(element, "register_hook") and callable(element.register_hook)) and \
            (hasattr(element, "requires_grad") and element.requires_grad)

    @staticmethod
    def is_recompute(call_stack=None):
        return is_recomputation(call_stack)

    @staticmethod
    def analyze_api_call_stack(name):
        try:
            call_stack = inspect.stack()
            if name.startswith("Primitive"):
                api_stack = call_stack[4:]
            else:
                api_stack = call_stack[5:]
        except Exception as e:
            logger.warning(f"The call stack of <{name}> failed to retrieve, {e}.")
            api_stack = None
            call_stack = None

        stack_str = []
        if api_stack:
            for (_, path, line, func, code, _) in api_stack:
                if not code:
                    continue
                if any(filter_path in path for filter_path in Const.STACK_FILTER_KEYWORDS) and \
                        Const.CALL_STACK_FLAG not in path:
                    continue
                stack_line = f"File {path}, line {str(line)}, in {func}, \n {code[0].strip()}"
                stack_str.append(stack_line)
        else:
            stack_str.append(Const.WITHOUT_CALL_STACK)
        is_recompute = PytorchDataProcessor.is_recompute(call_stack)
        del call_stack
        return tuple(stack_str), is_recompute

    @staticmethod
    def _analyze_torch_size(arg):
        return {"type": "torch.Size", "value": [int(x) for x in list(arg)]}

    @staticmethod
    def _analyze_memory_format(arg):
        # 获取内存格式
        format_type = PytorchDataProcessor.memory_format.get(arg)
        return {"type": "torch.memory_format", "format": format_type}

    @staticmethod
    def _analyze_process_group(arg):
        group_info = {"type": "torch.ProcessGroup"}
        try:
            group_ranks = dist.get_process_group_ranks(arg)
            group_info.update({"group_ranks": group_ranks})
            group_id = PytorchDataProcessor.process_group_hash(arg)
            group_info.update({"group_id": group_id})
        except Exception as e:
            logger.warning(f"Failed to get process group ranks info with error info: {e}.")
        return group_info

    @staticmethod
    def _analyze_reduce_op(arg):
        op_type = None
        try:
            op_type = str(arg)
        except Exception as e:
            logger.warning(f"Failed to get value of torch.distributed.ReduceOp with error info: {e}.")
        return {"type": "torch.distributed.ReduceOp", "value": op_type}

    @classmethod
    def get_special_types(cls):
        return super().get_special_types() + cls.pytorch_special_type

    def get_stat_info(self, data, async_dump=False, precision=Const.DUMP_PRECISION_LOW):
        tensor_stat = TensorStatInfo()
        if self.tensor_handler.is_empty_data(data):
            return tensor_stat
        data_clone = data.detach()
        if not data_clone.numel() or not data_clone.data_ptr():
            return tensor_stat
        if torch.is_complex(data_clone):
            if async_dump:
                logger.warning("Async dump do not support complex data!")
                return tensor_stat
            data_np = data_clone.cpu().numpy()
            data_abs = np.abs(data_np)
            tensor_stat.max = np.max(data_abs).item()
            tensor_stat.min = np.min(data_abs).item()
            tensor_stat.mean = np.mean(data_abs).item()
        elif data_clone.dtype == torch.bool:
            tensor_stat.max = torch.any(data_clone)
            tensor_stat.min = torch.all(data_clone)
        elif not data_clone.shape:
            tensor_stat.max = tensor_stat.min = tensor_stat.mean = tensor_stat.norm = data_clone.clone()
        else:
            if (precision == Const.DUMP_PRECISION_HIGH or data_clone.dtype == torch.float64
                    or not data_clone.is_floating_point()):
                data_clone = data_clone.float()
            tensor_stat.max = torch.max(data_clone)
            tensor_stat.min = torch.min(data_clone)
            tensor_stat.mean = torch.mean(data_clone)
            tensor_stat.norm = torch.norm(data_clone)
        return tensor_stat

    def dump_async_data(self):
        for file_path, tensor in self._async_dump_cache.items():
            self.tensor_handler.save_tensor(tensor, file_path)
        self._async_dump_cache.clear()

    def analyze_single_element(self, element, suffix_stack):
        if suffix_stack and suffix_stack[-1] in self.torch_object_key:
            return self.torch_object_key[suffix_stack[-1]](element)

        suffix_str = Const.SEP.join(str(s) for s in suffix_stack)
        type_analyzer = [
            (PytorchDataProcessor.builtin_type, self._analyze_builtin),
            (torch.Size, self._analyze_torch_size),
            (torch.Tensor, lambda e: self._analyze_tensor(e, suffix_str)),
            (torch.memory_format, self._analyze_memory_format),
            (dist.ProcessGroup, self._analyze_process_group),
            (dist.P2POp, lambda e: self._analyze_p2pop(e, suffix_str)),
            (dist.ReduceOp, self._analyze_reduce_op),
            (PytorchDataProcessor.np_type[:-1], self._analyze_numpy),
            (np.ndarray, lambda e: self._analyze_ndarray(e, suffix_str)),
        ]
        for type_key, analyze_fn in type_analyzer:
            if isinstance(element, type_key):
                return analyze_fn(element)
        return {}

    def _analyze_p2pop(self, arg, suffix):
        p2pop_info = {"class_type": "torch.distributed.P2POp"}
        try:
            tensor_info = self._analyze_tensor(arg.tensor, suffix)
            p2pop_info.update({"tensor": tensor_info})
            p2pop_info.update({"op": arg.op.__name__})
            p2pop_info.update({"peer": arg.peer})
            p2pop_info.update({"tag": arg.tag})
            group_id = PytorchDataProcessor.process_group_hash(
                arg.group) if arg.group else PytorchDataProcessor.process_group_hash(_get_default_group())
            p2pop_info.update({"group_id": group_id})
        except Exception as e:
            logger.warning(f"Failed to parse the P2POp content with error info: {e}.")
        return p2pop_info

    def _analyze_tensor(self, tensor, suffix):
        common_tensor = self.tensor_handler.convert_common_tensor(tensor)
        tensor_stat = self.get_stat_info(common_tensor, self.config.async_dump, self.config.precision)
        tensor_json = {}
        tensor_json.update({'type': self.tensor_handler.get_tensor_type(tensor)})
        tensor_json.update({'dtype': self.tensor_handler.get_tensor_dtype(tensor)})
        tensor_json.update({"shape": common_tensor.shape})

        stat_values = [
            tensor_stat.max,
            tensor_stat.min,
            tensor_stat.mean,
            tensor_stat.norm
        ]
        placeholder_index = self.data_writer.append_stat_to_buffer(stat_values)

        tensor_json.update({Const.TENSOR_STAT_INDEX: placeholder_index})
        tensor_json.update({"requires_grad": tensor.requires_grad})
        if self.tensor_handler.is_dtensor(tensor):
            dtensor_info = self.tensor_handler.get_dtensor_info(tensor)
            tensor_json.update(dtensor_info)

        if self.config.summary_mode == Const.MD5 and not self.config.async_dump:
            tensor_md5 = None
            if not self.tensor_handler.is_empty_data(tensor):
                t_cpu = common_tensor

                # 根据设备类型做同步，确保数据已准备好
                if t_cpu.device.type == "cuda":
                    t_cpu = t_cpu.to("cpu", non_blocking=True)
                    torch.cuda.synchronize()
                    # 先异步搬运再进行同步可以显著提升性能
                elif t_cpu.device.type == "npu":
                    t_cpu = t_cpu.to("cpu", non_blocking=True)
                    torch.npu.synchronize()
                t_cpu = t_cpu.detach()

                if self.config.task == Const.TENSOR and self.data_writer.bench_dump_file_path is not None:
                    tensor_md5 = PytorchDataProcessor.compute_crc32_from_tensor(t_cpu)
                    tensor_json.update({Const.MD5: tensor_md5})
                else:
                    if not t_cpu.is_contiguous():
                        t_cpu = t_cpu.contiguous()

                    future = self._crc_executor.submit(
                        PytorchDataProcessor.compute_crc32_from_tensor,
                        t_cpu
                    )

                    crc_placeholder = self.data_writer.append_crc32_to_buffer(future)
                    tensor_json[Const.MD5_INDEX] = crc_placeholder
            else:
                logger.debug(
                    "Calculating the md5 value of fake tensor or meta tensor is not supported, "
                    f"the current api/module name is {self.current_api_or_module_name}."
                )
                tensor_json.update({Const.MD5: tensor_md5})
        return tensor_json

    def _analyze_and_save_tensor(self, tensor, suffix):
        dump_data_name, file_path = self.get_save_file_path(suffix)
        single_arg = PytorchDataProcessor._analyze_tensor(self, tensor, suffix)
        common_tensor = self.tensor_handler.convert_common_tensor(tensor)
        if self.tensor_handler.is_empty_data(common_tensor):
            logger.debug(f"Saving fake tensor or meta tensor is not supported, the current tensor is {file_path}.")
            return single_arg
        if common_tensor.untyped_storage().data_ptr() == 0:
            logger.debug(f"Saving null-pointer tensor is not supported, the current tensor is {file_path}.")
            return single_arg

        single_arg.update({"data_name": dump_data_name})
        if self.config.async_dump:
            self._async_dump_cache[file_path] = common_tensor.clone().detach()
        else:
            self.tensor_handler.save_tensor(common_tensor, file_path)
        return single_arg

    def _analyze_and_save_ndarray(self, ndarray, suffix):
        dump_data_name, file_path = self.get_save_file_path(suffix)
        self.tensor_handler.save_tensor(torch.tensor(ndarray), file_path)
        ndarray_json = PytorchDataProcessor._analyze_ndarray(ndarray, suffix)
        ndarray_json.update({"data_name": dump_data_name})
        return ndarray_json


class StatisticsDataProcessor(PytorchDataProcessor):
    def _analyze_tensor(self, tensor, suffix):
        if any(item in self.current_api_or_module_name for item in self.config.tensor_list):
            return self._analyze_and_save_tensor(tensor, suffix)
        else:
            return super()._analyze_tensor(tensor, suffix)

    def _analyze_ndarray(self, ndarray, suffix):
        if any(item in self.current_api_or_module_name for item in self.config.tensor_list):
            return self._analyze_and_save_ndarray(ndarray, suffix)
        else:
            return super()._analyze_ndarray(ndarray, suffix)


class TensorDataProcessor(PytorchDataProcessor):
    def _analyze_tensor(self, tensor, suffix):
        return self._analyze_and_save_tensor(tensor, suffix)

    def _analyze_ndarray(self, ndarray, suffix):
        return self._analyze_and_save_ndarray(ndarray, suffix)


class DiffCheckDataProcessor(PytorchDataProcessor):
    __slots__ = [
        "cached_tensors_and_file_paths",
        "_bench_ref_path",
        "_bench_ref_mtime",
        "_bench_map",
        "_bench_state",  # 新增：按 API 的对比状态
    ]

    def __init__(self, config, data_writer):
        super().__init__(config, data_writer)
        self.has_diff = False

        self.cached_api_info = {}
        self.cached_tensors_and_file_paths = {}
        self.bits_for_diff = 8
        self.real_diff_nums = 0
        self.diff_nums = config.diff_nums

        # 新增：bench 基准缓存初始化
        self._bench_ref_path = None
        self._bench_ref_mtime = None
        self._bench_map = {}
        self._bench_state = {}  # key: api_name -> 状态字典

    @property
    def is_terminated(self):
        if self.diff_nums == -1:
            return False
        if self.real_diff_nums >= self.diff_nums:
            return True
        return False

    @staticmethod
    def _parse_data_name(data_name: str):
        """
        解析 data_name，例如：
        - "Functional.relu.2.forward.input.0.pt"
        - 兼容可选前缀 "name:" -> "name:Functional.relu.2.forward.input.0.pt"
        返回 (api, io, idx) 或 None
        """
        if not data_name:
            return None
        if data_name.startswith("name:"):
            data_name = data_name.split(":", 1)[1]

        # api 名本身可能包含若干个 '.'，所以用正则从右侧提取 io/idx/扩展名
        m = re.match(
            r"^(?=.{1,1024}$)(?P<api>.+)\.(?P<io>input|output)\.(?P<idx>\d+)\.\w+$",
            data_name
        )
        if not m:
            return None
        api = m.group("api")
        io = m.group("io")
        idx = int(m.group("idx"))
        return api, io, idx

    def analyze_forward_input(self, name, module, module_input_output: ModuleForwardInputsOutputs):
        self.has_diff = False

        self.cached_api_info = super().analyze_forward_input(name, module, module_input_output)
        return None

    def analyze_forward_output(self, name, module, module_input_output: ModuleForwardInputsOutputs):

        api_info_struct = super().analyze_forward_output(name, module, module_input_output)
        if name in self.cached_api_info and name in api_info_struct:
            self.cached_api_info[name].update(api_info_struct[name])
        elif name in api_info_struct:
            self.cached_api_info = api_info_struct
        self.handle_diff()
        return self.cached_api_info

    def analyze_forward(self, name, module, module_input_output: ModuleForwardInputsOutputs):
        self.has_diff = False

        api_info_struct = super().analyze_forward(name, module, module_input_output)
        self.handle_diff()
        return api_info_struct

    def analyze_backward(self, name, module, module_input_output: ModuleBackwardInputsOutputs):
        self.has_diff = False

        api_info_struct = super().analyze_backward(name, module, module_input_output)
        self.handle_diff()
        return api_info_struct

    def analyze_params(self, name, param_name, grad):
        self.has_diff = False

        api_info_struct = super().analyze_params(name, param_name, grad)
        self.handle_diff()
        return api_info_struct

    def handle_diff(self):
        if self.has_diff:
            for file_path, tensor in self.cached_tensors_and_file_paths.items():
                self.tensor_handler.save_tensor(tensor, file_path)
            self.real_diff_nums += 1
            if self.diff_nums != -1 and self.real_diff_nums >= self.diff_nums:
                logger.info(f"[{Const.TOOL_NAME}] Reached the preset diff times, "
                            f"current diff times: {self.real_diff_nums}.")
        api = getattr(self, "current_api_or_module_name", None)
        if api and api in self._bench_state:
            self._bench_state.pop(api, None)

        self.cached_tensors_and_file_paths = {}


    def _analyze_maybe_diff_flag(self):
        try:
            self.has_diff = torch_npu.npu.utils.get_npu_diff_flag()
            if self.has_diff:
                torch_npu.npu.utils.clear_npu_diff_flag()
        except Exception as e:
            logger.error(f"Diff check failed, the current environment may be abnormal.")
            raise RuntimeError(f"diff check failed") from e

    def _bench_expected_counts_for_api(self, api: str):
        """统计某 API 在 bench_map 里有多少个 Tensor 输入/输出"""
        n_in = n_out = 0
        for (a, io, _) in self._bench_map.keys():
            if a == api:
                if io == "input":
                    n_in += 1
                elif io == "output":
                    n_out += 1
        return n_in, n_out

    def _resolve_bench_json_path(self) -> str:
        p = getattr(self.data_writer, "bench_dump_file_path", None)
        if not p:
            return None
        p = os.path.join(p, "dump.json") if os.path.isdir(p) else p
        return p if os.path.isfile(p) else None

    def _ensure_bench_map_loaded(self) -> bool:
        """
        当路径变化或文件 mtime 变化时重载 dump.json，并构建 (api, 'input'/'output', idx) -> {md5, shape} 的索引。
        """
        path = self._resolve_bench_json_path()
        if not path:
            return False
        try:
            mtime = os.path.getmtime(path)
        except Exception as e:
            return False

        need_reload = (path != self._bench_ref_path) or (mtime != self._bench_ref_mtime)

        if need_reload:
            try:
                obj = load_json(path)
            except Exception as e:
                logger.warning(f"Failed to load bench dump.json: {e}")
                return False

            data = obj.get("data", {})
            self._bench_map = self._build_bench_map_from_json(data)
            self._bench_ref_path = path
            self._bench_ref_mtime = mtime

        return True

    def _build_bench_map_from_json(self, data: dict) -> dict:
        """
        data 结构：{ api_name: {input_args: [...], output: [...] } }
        只收集 Tensor 项：(api, io, idx) -> {"md5": str, "shape": list}
        """
        mp = {}
        total_inputs = 0
        total_outputs = 0
        for api_name, rec in data.items():
            ia = rec.get("input_args", [])
            oa = rec.get("output", [])
            # input_args
            input_count_this_api = 0
            for i, arg in enumerate(ia):
                if isinstance(arg, dict) and arg.get("type") == "torch.Tensor":
                    mp[(api_name, "input", i)] = {
                        "md5": arg.get("md5"),
                        "shape": arg.get("shape"),
                    }
                    input_count_this_api += 1
            total_inputs += input_count_this_api

            # output
            output_count_this_api = 0
            for i, out in enumerate(oa):
                if isinstance(out, dict) and out.get("type") == "torch.Tensor":
                    mp[(api_name, "output", i)] = {
                        "md5": out.get("md5"),
                        "shape": out.get("shape"),
                    }
                    output_count_this_api += 1
            total_outputs += output_count_this_api

        return mp

    def _analyze_maybe_diff_tensor(self, tensor_json):
        # 1) bench map 准备
        if not self._ensure_bench_map_loaded():
            return

        # 2) 解析 data_name -> (api, io, idx)
        data_name = tensor_json.get("data_name")
        parsed = self._parse_data_name(data_name)
        if not parsed:
            logger.debug(f"data_name parse failed: {data_name}")
            return
        api, io, idx = parsed

        # 3) 取/建 本 API 的状态
        st = self._bench_state.get(api)
        if st is None:
            n_in, _ = self._bench_expected_counts_for_api(api)
            st = {
                "expected_in": n_in,  # 标杆中该 API 期望的 Tensor 输入数
                "checked_in": 0,  # 已经校验过的“在标杆中存在的输入”个数
                "inputs_equal": True,  # 到目前为止，输入是否全部一致
                "seen_input_not_in_ref": False,  # 遇到“运行时存在但标杆里没有”的输入
                "any_output_neq": False,  # 是否发现过任一输出不一致（shape 同且 md5 不同）
            }
            self._bench_state[api] = st

        # 4) 找到标杆项
        ref = self._bench_map.get((api, io, idx))

        # 5) 当前 shape
        cur_shape = tensor_json.get("shape")
        if cur_shape is None:
            return
        try:
            cur_shape = list(cur_shape)
        except Exception as e:
            logger.warning("[BENCH]", "shape to list failed:", repr(e), "-> skip")
            return

        # 6) 输入与输出分别处理
        if io == "input":
            # —— 输入阶段：只维护“输入是否一致”的状态 —— #
            if ref is None:
                # 运行时有输入，但标杆里没有对应条目 => 不能断言“输入一致”
                st["inputs_equal"] = False
                st["seen_input_not_in_ref"] = True

                return

            ref_shape = ref.get("shape")
            ref_md5 = ref.get("md5")

            # 标杆有该输入，计入已校验
            st["checked_in"] += 1

            # shape 必须一致
            if list(ref_shape) != list(cur_shape):
                st["inputs_equal"] = False

                return

            # 取当前 md5
            cur_md5 = tensor_json.get(Const.MD5) if Const.MD5 in tensor_json else tensor_json.get("md5")

            if cur_md5 is None or ref_md5 is None:
                # 缺少 md5 信息，无法断言一致
                st["inputs_equal"] = False
                return

            # md5 必须一致
            if str(cur_md5) != str(ref_md5):
                st["inputs_equal"] = False
            return  # 输入阶段不触发 has_diff

        else:  # io == "output"
            # —— 输出阶段：仅当“所有输入一致且已校验完所有输入”时，才检查输出不一致以置位 —— #
            # 若标杆无此输出，按照你的规则：不能断言输出不一致，直接跳过
            if ref is None:
                return

            ref_shape = ref.get("shape")
            ref_md5 = ref.get("md5")

            # shape 必须一致才比较 md5
            if list(ref_shape) != list(cur_shape):
                return

            cur_md5 = tensor_json.get(Const.MD5) if Const.MD5 in tensor_json else tensor_json.get("md5")
            if cur_md5 is None or ref_md5 is None:
                return

            # 只有当“输入全部一致且已校验完所有输入”时，才允许判定输出不一致
            inputs_ok = (
                    st["inputs_equal"]
                    and (st["checked_in"] == st["expected_in"])
                    and (not st["seen_input_not_in_ref"])
            )

            if inputs_ok and (str(cur_md5) != str(ref_md5)):
                st["any_output_neq"] = True
                self.has_diff = True

    def _analyze_tensor(self, tensor, suffix):
        dump_data_name, file_path = self.get_save_file_path(suffix)
        self.cached_tensors_and_file_paths.update({file_path: tensor})
        single_arg = super()._analyze_tensor(tensor, suffix)
        single_arg.update({"data_name": dump_data_name})
        if not self.has_diff:
            self._analyze_maybe_diff_tensor(single_arg)
        return single_arg


class KernelDumpDataProcessor(PytorchDataProcessor):
    def __init__(self, config, data_writer):
        super().__init__(config, data_writer)
        self.enable_kernel_dump = True
        self.is_found_output_tensor = False
        self.is_found_grad_input_tensor = False
        self.forward_args = None
        self.forward_kwargs = None
        self.forward_output_tensor = None
        self.grad_input_tensor = None

    @staticmethod
    def start_kernel_dump(config_path):
        torch_npu.npu.synchronize()
        torch_npu.npu.init_dump()
        torch_npu.npu.set_dump(config_path)
        torch_npu.npu.synchronize()

    @staticmethod
    def stop_kernel_dump():
        torch_npu.npu.synchronize()
        torch_npu.npu.finalize_dump()
        torch_npu.npu.synchronize()

    @staticmethod
    def _print_unsupported_log(api_name):
        logger.warning(f"The kernel dump does not support the {api_name} API.")

    def analyze_forward_input(self, name, module, module_input_output):
        if not self.enable_kernel_dump:
            return
        if is_gpu:
            logger.warning("The current environment is not a complete NPU environment, and kernel dump cannot be used.")
            self.enable_kernel_dump = False
            return

        if self.config.is_backward_kernel_dump:
            try:
                self.forward_args = self.clone_and_detach_tensor(module_input_output.args)
                self.forward_kwargs = self.clone_and_detach_tensor(module_input_output.kwargs)
                output = module.forward(*self.forward_args, **self.forward_kwargs)
            except Exception as e:
                if isinstance(e, MsprobeException):
                    logger.warning(str(e))
                self._print_unsupported_log(name)
                self.enable_kernel_dump = False
                return

            self.analyze_element(convert_tuple(output))
            if not self.is_found_output_tensor:
                self._print_unsupported_log(name)
                self.enable_kernel_dump = False
            return
        self.start_kernel_dump(self.config.kernel_config_path)

    def analyze_forward_output(self, name, module, module_input_output):
        if not self.enable_kernel_dump:
            return
        if self.config.is_backward_kernel_dump:
            return
        self.enable_kernel_dump = False
        self.stop_kernel_dump()
        logger.info(f"The kernel data of {name} is dumped successfully.")

    def analyze_backward(self, name, module, module_input_output):
        if not self.enable_kernel_dump:
            return
        self.enable_kernel_dump = False

        self.analyze_element(module_input_output.grad_input)
        if not self.is_found_grad_input_tensor:
            self._print_unsupported_log(name)
            return
        self.start_kernel_dump(self.config.kernel_config_path)

        try:
            self.forward_output_tensor.backward(self.grad_input_tensor, retain_graph=True)
        except Exception:
            self._print_unsupported_log(name)
            self.stop_kernel_dump()
            return

        self.stop_kernel_dump()
        logger.info(f"The kernel data of {name} is dumped successfully.")

    @recursion_depth_decorator(
        "KernelDump: KernelDumpDataProcessor.clone_and_detach_tensor",
        max_depth=Const.DUMP_MAX_DEPTH
    )
    def clone_and_detach_tensor(self, input_params):
        if isinstance(input_params, torch.Tensor):
            if is_float8_tensor(input_params):
                raise MsprobeException(
                    MsprobeException.UNSUPPORTED_TYPE_ERROR,
                    f"L2 backward dump does not support float8 type."
                )
            if input_params.requires_grad:
                return input_params.clone().detach().requires_grad_()
            return input_params.clone()
        elif isinstance(input_params, tuple):
            return tuple(self.clone_and_detach_tensor(x) for x in input_params)
        elif isinstance(input_params, list):
            return list(self.clone_and_detach_tensor(x) for x in input_params)
        elif isinstance(input_params, dict):
            return {k: self.clone_and_detach_tensor(v) for k, v in input_params.items()}
        else:
            return input_params

    def analyze_single_element(self, element, suffix_stack):
        if is_float8_tensor(element):
            return {}
        if isinstance(element, torch.Tensor):
            if not self.is_found_output_tensor:
                if element.requires_grad:
                    self.forward_output_tensor = element
                    self.is_found_output_tensor = True
                return {}
            if not self.is_found_grad_input_tensor:
                self.grad_input_tensor = element.clone()
                self.is_found_grad_input_tensor = True
        return {}

    def reset_status(self):
        self.enable_kernel_dump = True
        self.is_found_output_tensor = False
        self.is_found_grad_input_tensor = False
        self.forward_args = None
        self.forward_kwargs = None
        self.forward_output_tensor = None
        self.grad_input_tensor = None



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import collections
import functools
import inspect
import os
import re
import threading
import time
import importlib.util
from collections import OrderedDict
from datetime import datetime, timezone

import numpy as np

from msprobe.core.common.const import Const, CompareConst
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.core.common.exceptions import MsprobeException
from msprobe.core.common.file_utils import (FileOpen, check_file_or_directory_path, load_json, load_construct_json)
from msprobe.core.common.log import logger

device = collections.namedtuple('device', ['type', 'index'])
prefixes = ['api_stack', 'list', 'range', 'acl']
file_suffix_to_file_type = {
    "dump.json": Const.DUMP_JSON_FILE,
    "debug.json": Const.DEBUG_JSON_FILE,
}


class MsprobeBaseException(Exception):
    """
    Base class for all custom exceptions.
    """
    # 所有的错误代码
    NONE_ERROR = 0
    INVALID_PATH_ERROR = 1
    OPEN_FILE_ERROR = 2
    CLOSE_FILE_ERROR = 3
    READ_FILE_ERROR = 4
    WRITE_FILE_ERROR = 5
    INVALID_FILE_ERROR = 6
    PERMISSION_ERROR = 7
    INDEX_OUT_OF_BOUNDS_ERROR = 8
    NO_DUMP_FILE_ERROR = 9
    INVALID_DATA_ERROR = 10
    INVALID_PARAM_ERROR = 11
    INVALID_DUMP_RATIO = 12
    INVALID_DUMP_FILE = 13
    UNKNOWN_ERROR = 14
    INVALID_DUMP_MODE = 15
    PARSE_FILE_ERROR = 16
    INVALID_COMPARE_MODE = 17
    OVER_SIZE_FILE_ERROR = 18
    INVALID_SUMMARY_MODE = 19
    INVALID_TASK_ERROR = 20
    DETACH_ERROR = 21
    INVALID_OBJECT_TYPE_ERROR = 22
    INVALID_CHAR_ERROR = 23
    RECURSION_LIMIT_ERROR = 24
    INVALID_ATTRIBUTE_ERROR = 25
    OUTPUT_HOOK_ERROR = 26
    INPUT_HOOK_ERROR = 27
    FUNCTION_CALL_ERROR = 28
    FORWARD_DATA_COLLECTION_ERROR = 29
    BACKWARD_DATA_COLLECTION_ERROR = 30
    INVALID_KEY_ERROR = 31
    MISSING_HEADER_ERROR = 32
    MERGE_COMPARE_RESULT_ERROR = 33
    NAMES_STRUCTS_MATCH_ERROR = 34
    INVALID_STATE_ERROR = 35
    INVALID_API_NAME_ERROR = 36
    CROSS_FRAME_ERROR = 37
    MISSING_THRESHOLD_ERROR = 38
    WRONG_THRESHOLD_ERROR = 39
    MULTIPROCESS_ERROR = 40

    def __init__(self, code, error_info: str = ""):
        super(MsprobeBaseException, self).__init__()
        self.code = code
        self.error_info = error_info

    def __str__(self):
        return self.error_info


class CompareException(MsprobeBaseException):
    """
    Class for Accuracy Compare Exception
    """

    def __init__(self, code, error_info: str = ""):
        super(CompareException, self).__init__(code, error_info)


class DumpException(MsprobeBaseException):
    """
    Class for Dump Exception
    """

    def __init__(self, code, error_info: str = ""):
        super(DumpException, self).__init__(code, error_info)

    def __str__(self):
        return f"Dump Error Code {self.code}: {self.error_info}"


class ThreadSafe:
    """
    线程安全控制工具类，提供三种使用方式：
    1.上下文管理器：with ThreadSafe()
    2.主动加锁与释放锁：ThreadSafe.acquire()/ThreadSafe.release()
    3.方法装饰器：@ThreadSafe.synchronized
    """
    _lock = threading.RLock()

    def __enter__(self):
        self.__class__._lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__class__._lock.release()

    @classmethod
    def acquire(cls):
        cls._lock.acquire()

    @classmethod
    def release(cls):
        cls._lock.release()

    @classmethod
    def synchronized(cls, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with cls._lock:
                return func(*args, **kwargs)

        return wrapper


class ModuleQueue:
    def __init__(self):
        self.queue = OrderedDict()

    def add_name(self, name):
        self.queue[name] = True

    def remove_name(self, name):
        if name in self.queue:
            del self.queue[name]

    def find_last(self, name):
        """
        在队列中找到当前 Module/Cell 的父节点名称并返回，若找不到则返回None

        Args:
            name: 需要寻找父节点的 Module/Cell 的名称

        Returns:
            返回父节点名称，找不到则返回None

        Examples:
            父节点名称格式: Module.module1.module1.forward.0
            子节点名称格式: Module.module1.module2.Module2.forward.0
            匹配关系: Module/Cell 的名称总能被点(.)分割符分成5个部分及以上，子节点截断后4个点和父节点截断后3个点的前缀名称是匹配的
        """
        child_parts = name.split('.')
        if len(child_parts) < 5:
            return None
        child_name_prefix = '.'.join(child_parts[:-4])
        if child_name_prefix in Const.MODULE_PREFIX:
            return None

        for parent_name in reversed(self.queue):
            parent_parts = parent_name.split('.')
            if len(parent_parts) < 5:
                return None
            parent_name_prefix = '.'.join(parent_parts[:-3])
            if parent_name_prefix == child_name_prefix:
                return parent_name
        return None


def is_json_file(file_path):
    if isinstance(file_path, str) and file_path.lower().endswith('.json'):
        return True
    else:
        return False


def check_compare_param(input_param, output_path, dump_mode, stack_mode):
    if not isinstance(input_param, dict):
        logger.error(f"Invalid input parameter 'input_param', the expected type dict but got {type(input_param)}.")
        raise CompareException(CompareException.INVALID_PARAM_ERROR)
    if not isinstance(output_path, str):
        logger.error(f"Invalid input parameter 'output_path', the expected type str but got {type(output_path)}.")
        raise CompareException(CompareException.INVALID_PARAM_ERROR)

    def check_json_path(json_path_str):
        json_path = input_param.get(json_path_str)
        check_file_or_directory_path(json_path, False)
        json_type_check = is_json_file(json_path)
        if not json_type_check:
            logger.error(f"Invalid {json_path_str}: {json_path}, please check!")
            raise CompareException(CompareException.INVALID_PATH_ERROR)

    check_json_path("npu_path")
    check_json_path("bench_path")
    if stack_mode:
        check_json_path("stack_path")

    if dump_mode == Const.ALL:
        check_file_or_directory_path(input_param.get("npu_dump_data_dir"), True)
        check_file_or_directory_path(input_param.get("bench_dump_data_dir"), True)
    check_file_or_directory_path(output_path, True)

    with FileOpen(input_param.get("npu_path"), "r") as npu_json, \
            FileOpen(input_param.get("bench_path"), "r") as bench_json:
        _check_json(npu_json, input_param.get("npu_path"))
        _check_json(bench_json, input_param.get("bench_path"))
    if stack_mode:
        with FileOpen(input_param.get("stack_path"), "r") as stack_json:
            _check_json(stack_json, input_param.get("stack_path"))


def _check_json(json_file_handle, file_name):
    tensor_line = json_file_handle.readline()
    if not tensor_line:
        logger.error("dump file {} have empty line!".format(file_name))
        raise CompareException(CompareException.INVALID_DUMP_FILE)
    json_file_handle.seek(0, 0)


def check_json_file(input_param, npu_json, bench_json, stack_json):
    _check_json(npu_json, input_param.get("npu_path"))
    _check_json(bench_json, input_param.get("bench_path"))
    _check_json(stack_json, input_param.get("stack_path"))


def check_regex_prefix_format_valid(prefix):
    """
        validate the format of the regex prefix

    Args:
        prefix (str): The prefix string to validate.

    Returns:
        no returns

    Raises:
        ValueError: if the prefix length exceeds Const.REGEX_PREFIX_MAX_LENGTH characters or the prefix do not match
        the given pattern Const.REGEX_PREFIX_PATTERN
    """
    if len(prefix) > Const.REGEX_PREFIX_MAX_LENGTH:
        raise ValueError(f"Maximum length of prefix is {Const.REGEX_PREFIX_MAX_LENGTH}, while current length "
                         f"is {len(prefix)}")
    if not re.match(Const.REGEX_PREFIX_PATTERN, prefix):
        raise ValueError(f"prefix contains invalid characters, prefix pattern {Const.REGEX_PREFIX_PATTERN}")


def add_time_as_suffix(name):
    return '{}_{}.csv'.format(name, time.strftime("%Y%m%d%H%M%S", time.localtime(time.time())))


def add_time_with_xlsx(name):
    return '{}_{}.xlsx'.format(name, time.strftime("%Y%m%d%H%M%S", time.localtime(time.time())))


def add_time_with_json(name):
    return '{}_{}.json'.format(name, time.strftime("%Y%m%d%H%M%S", time.localtime(time.time())))


def add_time_with_yaml(name):
    return '{}_{}.yaml'.format(name, time.strftime("%Y%m%d%H%M%S", time.localtime(time.time())))


def get_time():
    return datetime.now(tz=timezone.utc).strftime("%Y%m%d_%H%M%S")


def format_value(value):
    return float('{:.12f}'.format(value))


@recursion_depth_decorator('msprobe.core.common.utils.md5_find', max_depth=Const.DUMP_MAX_DEPTH)
def md5_find(data, json_type=Const.DUMP_JSON_FILE):
    if json_type == Const.DUMP_JSON_FILE:
        for key_op in data:
            for api_info in data[key_op]:
                if isinstance(data[key_op][api_info], list):
                    for data_detail in data[key_op][api_info]:
                        if data_detail and Const.MD5 in data_detail:
                            return True
                if isinstance(data[key_op][api_info], bool):
                    continue
                elif data[key_op][api_info] and Const.MD5 in data[key_op][api_info]:
                    return True
    elif json_type == Const.DEBUG_JSON_FILE:
        if isinstance(data, dict):
            if Const.MD5 in data:
                return True
            else:
                for _, data_info in data.items():
                    if md5_find(data_info, Const.DEBUG_JSON_FILE):
                        return True
        elif isinstance(data, list):
            for data_info in data:
                if md5_find(data_info, Const.DEBUG_JSON_FILE):
                    return True
        else:
            return False
    return False


def detect_framework_by_dump_json(file_path):
    json_data = load_json(file_path)
    framework = json_data.get("framework", None)
    if framework in [Const.PT_FRAMEWORK, Const.MS_FRAMEWORK]:
        return framework
    pattern_ms = r'"type":\s*"mindspore'
    pattern_pt = r'"type":\s*"torch'
    with FileOpen(file_path, 'r') as file:
        for line in file:
            if re.search(pattern_ms, line):
                return Const.MS_FRAMEWORK
            if re.search(pattern_pt, line):
                return Const.PT_FRAMEWORK
    logger.error(f"{file_path} must be based on the MindSpore or PyTorch framework.")
    raise CompareException(CompareException.INVALID_PARAM_ERROR)


def get_stack_construct_by_dump_json_path(dump_json_path):
    if not dump_json_path:
        logger.error("The path is empty. Please enter a valid path.")
        raise CompareException(CompareException.INVALID_PATH_ERROR)
    directory = os.path.dirname(dump_json_path)
    check_file_or_directory_path(directory, True)
    stack_json = os.path.join(directory, "stack.json")
    construct_json = os.path.join(directory, "construct.json")

    stack_json_exist = os.path.exists(stack_json)
    construct_json_exist = os.path.exists(construct_json)

    if not stack_json_exist and not construct_json_exist:
        logger.info("stack.json and construct.json not found")
        return {}, {}
    if not stack_json_exist or not construct_json_exist:
        logger.error("stack.json or construct.json not found, please check.")
        raise CompareException(CompareException.INVALID_PATH_ERROR)

    stack = load_json(stack_json)
    construct, _ = load_construct_json(construct_json)
    return stack, construct


def set_dump_path(input_param):
    npu_path = input_param.get("npu_path", None)
    bench_path = input_param.get("bench_path", None)
    dump_json_path_valid = npu_path is not None and npu_path.endswith("dump.json") and \
                           bench_path is not None and bench_path.endswith("dump.json")
    debug_json_path_valid = npu_path is not None and npu_path.endswith("debug.json") and \
                            bench_path is not None and bench_path.endswith("debug.json")
    if not dump_json_path_valid and not debug_json_path_valid:
        logger.error(f"Please check the json path is valid and ensure that neither npu_path nor bench_path is None.")
        raise CompareException(CompareException.INVALID_PATH_ERROR)
    input_param[CompareConst.NPU_DUMP_DATA_DIR] = os.path.join(os.path.dirname(npu_path), Const.DUMP_TENSOR_DATA)
    input_param[CompareConst.BENCH_DUMP_DATA_DIR] = os.path.join(os.path.dirname(bench_path), Const.DUMP_TENSOR_DATA)


def get_file_type(file_path):
    if not isinstance(file_path, str):
        logger.error("get_file_type failed, check the type of file_path.")
        raise CompareException(CompareException.INVALID_PATH_ERROR)
    file_type = file_suffix_to_file_type.get(file_path.split(Const.SCOPE_SEPARATOR)[-1])
    if file_type is None:
        logger.error("get_file_type failed, file_path is neither dump.json nor debug.json.")
        raise CompareException(CompareException.INVALID_PATH_ERROR)
    return file_type


def check_dump_json_key(json_data, device_type):
    task = json_data.get('task', None)
    if not task:
        logger.error(f"Task for {device_type} is empty, please check.")
        raise CompareException(CompareException.INVALID_TASK_ERROR)
    if 'data' not in json_data:
        logger.error(f"Missing 'data' in dump.json, please check dump.json of {device_type}.")
        raise CompareException(CompareException.INVALID_DATA_ERROR)
    api_data = json_data.get('data')
    if not isinstance(api_data, dict):
        logger.error(f"Invalid type for 'data': expected a dict. Please check dump.json of {device_type}.")
        raise CompareException(CompareException.INVALID_DATA_ERROR)
    return task, api_data


def get_dump_mode(input_param):
    npu_path = input_param.get("npu_path", None)
    bench_path = input_param.get("bench_path", None)
    npu_json_data = load_json(npu_path)
    bench_json_data = load_json(bench_path)
    json_type = get_file_type(file_path=npu_path)

    npu_task, npu_api_data = check_dump_json_key(npu_json_data, 'npu')
    bench_task, bench_api_data = check_dump_json_key(bench_json_data, 'bench')

    if npu_task != bench_task:
        logger.error(f"Please check the dump task is consistent.")
        raise CompareException(CompareException.INVALID_TASK_ERROR)

    if npu_task == Const.TENSOR:
        return Const.ALL

    if npu_task == Const.STRUCTURE:
        return Const.STRUCTURE

    if npu_task == Const.STATISTICS:
        npu_md5_compare = md5_find(npu_api_data, json_type)
        bench_md5_compare = md5_find(bench_api_data, json_type)
        if npu_md5_compare == bench_md5_compare:
            return Const.MD5 if npu_md5_compare else Const.SUMMARY
        else:
            logger.error(f"Please check the dump task is consistent, "
                         f"dump mode of npu and bench should both be statistics or md5.")
            raise CompareException(CompareException.INVALID_TASK_ERROR)

    logger.error(f"Compare applies only to task is tensor or statistics")
    raise CompareException(CompareException.INVALID_TASK_ERROR)


def get_header_index(header_name, dump_mode):
    header = CompareConst.HEAD_OF_COMPARE_MODE.get(dump_mode)
    if not header:
        logger.error(f"{dump_mode} not in {CompareConst.HEAD_OF_COMPARE_MODE}")
        raise CompareException(CompareException.INVALID_PARAM_ERROR)
    if header_name not in header:
        logger.error(f"{header_name} not in data name")
        raise CompareException(CompareException.INVALID_PARAM_ERROR)
    return header.index(header_name)


def convert_tuple(data):
    return data if isinstance(data, tuple) else (data,)


def check_op_str_pattern_valid(string, op_name=None, stack=False):
    if isinstance(string, str) and is_invalid_pattern(string):
        if stack:
            message = f"stack info of {op_name} contains special characters, please check!"
        elif not op_name:
            message = f"api name contains special characters, please check!"
        else:
            message = f"data info of {op_name} contains special characters, please check!"
        logger.error(message)
        raise CompareException(CompareException.INVALID_CHAR_ERROR)


def is_invalid_pattern(string):
    pattern = Const.STRING_BLACKLIST
    return re.search(pattern, string)


def is_int(x):
    return isinstance(x, int) and not isinstance(x, bool)


def print_tools_ends_info():
    total_len = len(Const.TOOL_ENDS_SUCCESSFULLY) + Const.FILL_CHAR_NUMS
    logger.info('*' * total_len)
    logger.info(f"*{Const.TOOL_ENDS_SUCCESSFULLY.center(total_len - 2)}*")
    logger.info('*' * total_len)


def get_step_or_rank_from_string(step_or_rank, obj):
    splited = step_or_rank.split(Const.HYPHEN)
    if len(splited) == 2:
        try:
            borderlines = int(splited[0]), int(splited[1])
        except (ValueError, IndexError) as e:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   "The hyphen(-) must start and end with decimal numbers.") from e
    else:
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                               f'The string parameter for {obj} only supports formats like "3-5". '
                               f'Now string parameter for {obj} is "{step_or_rank}".')
    if all(Const.STEP_RANK_MINIMUM_VALUE <= b <= Const.STEP_RANK_MAXIMUM_VALUE for b in borderlines):
        if borderlines[0] <= borderlines[1]:
            continual_step_or_rank = list(range(borderlines[0], borderlines[1] + 1))
        else:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f'For the hyphen(-) in {obj}, the left boundary ({borderlines[0]}) cannot be '
                                   f'greater than the right boundary ({borderlines[1]}).')
    else:
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                               f"The boundaries must fall within the range of "
                               f"[{Const.STEP_RANK_MINIMUM_VALUE}, {Const.STEP_RANK_MAXIMUM_VALUE}].")
    return continual_step_or_rank


def get_real_step_or_rank(step_or_rank_input, obj):
    if obj not in [Const.STEP, Const.RANK]:
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                               f"Only support parsing {[Const.STEP, Const.RANK]}, the current parsing object is {obj}.")
    if step_or_rank_input is None:
        return []
    if not isinstance(step_or_rank_input, list):
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR, f"{obj} is invalid, it should be a list")
    if len(step_or_rank_input) > Const.STEP_RANK_MAXIMUM_VALUE:
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                               f"{obj} is invalid, its length cannot exceed {Const.STEP_RANK_MAXIMUM_VALUE}")

    real_step_or_rank = []
    for element in step_or_rank_input:
        if not is_int(element) and not isinstance(element, str):
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"{obj} element {element} must be an integer or string.")
        if is_int(element):
            if not Const.STEP_RANK_MINIMUM_VALUE <= element <= Const.STEP_RANK_MAXIMUM_VALUE:
                raise MsprobeException(
                    MsprobeException.INVALID_PARAM_ERROR,
                    f"Each element of {obj} must be between {Const.STEP_RANK_MINIMUM_VALUE} and "
                    f"{Const.STEP_RANK_MAXIMUM_VALUE}, currently it is {element}."
                )
            real_step_or_rank.append(element)
            continue
        continual_step_or_rank = get_step_or_rank_from_string(element, obj)
        real_step_or_rank.extend(continual_step_or_rank)
    real_step_or_rank = list(set(real_step_or_rank))
    real_step_or_rank.sort()
    return real_step_or_rank


def check_init_step(step):
    if not is_int(step):
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                               f"{step} must be an integer")
    if not step >= 0:
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                               f"{step} must be greater than or equal to 0")


def check_token_range(token_range):
    if token_range is None:
        return
    if not isinstance(token_range, (list, tuple)):
        logger.error("Token_range must be a list or tuple.")
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if len(token_range) != 2:
        logger.error("Token_range must contains exactly 2 elements.")
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

    start, end = token_range
    if not is_int(start) or not is_int(end):
        logger.error("Start and end in token_range must be integer.")
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if start > end:
        logger.error("Start in token_range must less than the end.")
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if start < 0:
        logger.error("Start in token_range must >= 0.")
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)


def check_seed_all(seed, mode, rm_dropout):
    if is_int(seed):
        if seed < 0 or seed > Const.MAX_SEED_VALUE:
            logger.error(f"Seed must be between 0 and {Const.MAX_SEED_VALUE}.")
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    else:
        logger.error("Seed must be integer.")
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if not isinstance(mode, bool):
        logger.error("seed_all mode must be bool.")
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if not isinstance(rm_dropout, bool):
        logger.error("The rm_dropout parameter must be bool.")
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)


def safe_get_value(container, index, container_name, key=None):
    try:
        # 处理字典情况
        if isinstance(container, dict):
            return container.get(key)[index]
        # 处理列表、元组、numpy情况
        elif isinstance(container, (list, tuple, np.ndarray)):
            return container[index]
        else:
            err_msg = f"Unsupported container type for '{container_name}': {type(container)}"
            logger.error(err_msg)
            raise MsprobeBaseException(MsprobeBaseException.INVALID_OBJECT_TYPE_ERROR)
    except IndexError as e:
        err_msg = "index out of bounds error occurs, please check!\n" \
                  f"{container_name} is {container}\n" \
                  f"index is {index}"
        logger.error(err_msg)
        raise MsprobeBaseException(MsprobeBaseException.INDEX_OUT_OF_BOUNDS_ERROR) from e
    except TypeError as e:
        err_msg = "wrong type, please check!\n" \
                  f"{container_name} is {container}\n" \
                  f"index is {index}\n" \
                  f"key is {key}"
        logger.error(err_msg)
        raise MsprobeBaseException(MsprobeBaseException.INVALID_OBJECT_TYPE_ERROR) from e


def check_str_param(param):
    if not re.match(Const.REGEX_PREFIX_PATTERN, param):
        logger.error('The parameter {} contains special characters.'.format(param))
        raise MsprobeBaseException(MsprobeBaseException.INVALID_CHAR_ERROR)


class DumpPathAggregation:
    dump_file_path = None
    stack_file_path = None
    construct_file_path = None
    dump_tensor_data_dir = None
    debug_file_path = None


def is_save_variable_valid(variable, valid_special_types, depth=0):
    if depth > Const.DUMP_MAX_DEPTH:
        return False
    if isinstance(variable, valid_special_types):
        return True
    elif isinstance(variable, (list, tuple)):
        return all(is_save_variable_valid(item, valid_special_types, depth + 1) for item in variable)
    elif isinstance(variable, dict):
        return all(isinstance(key, str) and is_save_variable_valid(value, valid_special_types, depth + 1)
                   for key, value in variable.items())
    else:
        return False


def replace_last_occurrence(text, old, new):
    if text is None:
        return text
    index = text.rfind(old)
    if index != -1:
        return text[:index] + text[index:].replace(old, new, 1)
    return text


def load_stack_json(stack_path):
    stack_dict = load_json(stack_path)

    if not isinstance(stack_dict, dict):
        raise MsprobeException(
            MsprobeException.INVALID_PARAM_ERROR,
            "The format of the stack.json is incorrect, the outermost layer of stack.json should be a dict type."
        )

    if not stack_dict.get(Const.NEW_STACK_FLAG):
        return stack_dict

    new_stack_dict = {}
    for stack_info in stack_dict.values():
        if not isinstance(stack_info, list) or len(stack_info) != 2:
            continue

        api_list, stack_str = stack_info
        if not isinstance(api_list, list):
            continue

        for api_name in api_list:
            new_stack_dict.update({api_name: stack_str})
    return new_stack_dict


def analyze_api_call_stack(name):
    try:
        api_stack = inspect.stack()[2:]
    except Exception as e:
        logger.warning(f"The call stack of {name} failed to retrieve, {e}.")
        api_stack = None
    stack_str = []
    if api_stack:
        for (_, path, line, func, code, _) in api_stack:
            if not code:
                continue
            stack_line = f"File {path}, line {str(line)}, in {func}, \n {code[0].strip()} \n"
            stack_str.append(stack_line)
    else:
        stack_str.append(Const.WITHOUT_CALL_STACK)
    return "".join(stack_str)


def check_extern_input_list(input_list):
    if not isinstance(input_list, list):
        raise Exception("input is not a list")
    if len(input_list) > Const.EXTERN_INPUT_LIST_MAX_LEN:
        raise Exception(f"input list exceed max length {Const.EXTERN_INPUT_LIST_MAX_LEN}")


def check_process_num(process_num):
    if not is_int(process_num) or process_num <= 0:
        raise ValueError(f"process_num({process_num}) is not a positive integer")
    if process_num > Const.MAX_PROCESS_NUM:
        raise ValueError(f"The maximum supported process_num is {Const.MAX_PROCESS_NUM}, current value: {process_num}.")


def is_module_available(module_name):
    spec = importlib.util.find_spec(module_name)
    return spec is not None


def check_rank_id(rank_id):
    if rank_id is None:
        return
    if not is_int(rank_id):
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                               f"{rank_id} must be an integer.")
    if not rank_id >= 0:
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                               f"{rank_id} must be greater than or equal to 0.")



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

# 语言切换
ZH_CN = "zh-CN"
EN = "en"
language = {
    ZH_CN: {
        'typeError': '文件类型错误',
        'searchTypeError': '搜索类型错误',
        'nodeInfoError': '查询参数nodeInfo及其子项必须存在且必须为字典',
        'metaDataError': '查询参数“metaData”及其子项必须存在且必须为字典',
        'dbInitError': '数据库初始化失败',
        'loadConfigError': '加载图配置信息失败',
        'rankStepNullError': 'rank或step或micro_step为空',
        'loadNodeError': '加载图所有节点列表失败',
        'expandNodeError': '节点展开或收起发生错误',
        'searchNodeError': '节点搜索发生错误',
        'graphTypeError': '图类型错误,仅支持 "NPU" 或 "BENCH"',
        'getNodeInfoError': '获取节点信息失败',
        'taskTypeError': '任务类型不支持',
        'matchNodeError': '匹配节点失败',
        'loadConfigFileError': '读取配置文件失败',
        'matchNodeLinksNullError': '匹配配置文件为空',
        'parameterError': '参数错误',   
        'nodesTypeError': '节点类型不一致, 无法添加匹配关系',
        'taskTypeWrong': 'task类型错误',
        'nodeNotExist': '节点不存在',
        'ioEmptyError': '输入或输出统计误差值为空',
        'outputEmptyError': '输出统计误差值为空，计算精度误差失败',
        'operationFailed': '操作失败：',
        'or': '或',
        'nodeUnmatchError': '节点未匹配，请先匹配节点',
        'deleteMatchNodeFailed': '删除匹配节点失败',
        'noUpdateNodesFound': '未找到可更新的节点',
        'updateNodeAccuracyFailed': '更新节点精度失败',
        'updateDatabaseFailed': '更新数据库失败',
        'updateColorFailed': '更新颜色失败',
        'saveMatchRelationshipFailed': '保存匹配关系失败',
        'selectedNodeCantMatched': '选择的节点不可匹配',
        'loadErrorBySecurityIssue': '文件存在安全问题，读取文件失败',
        'loadConfigFailed': '获取配置信息失败, 请检查目录中第一个文件',
        'debug': '调试侧',
        'bench': '标杆侧',
        'obtainNodesAccuracyFailed': '获取符合精度误差节点失败',
        'notSupportOverflow': '多图模式下不支持溢出检测',
        'obtainNodesOverflowFailed': '获取符合溢出检测节点失败',
        'updatePrecisionFailed': '更新精度误差失败',
        'nodeTypeError': '节点类型错误',
        'readFileFailed': '读取文件失败',
        'parameterEmpty': '参数为空',
        'saveFileFailed': '保存文件失败',
        'dbConncetionNotInit': '数据库链接未初始化',
        'parseJsonFailed': '解析JSON失败',
        'getNodeListFailed': '获取节点列表失败',
        'initGraphJsonFailed': '初始化json图数据失败',
        'colorsNotObject': 'colors 必须是一个对象',
        'colotsEmpty': 'colors 不能为空',
        'illegalColorKey': '非法颜色键: ',
        'colorValuesNotObject': '颜色值必须是对象: ',
        'missingValueField': '缺少 value 字段: ',
        'notArrayOfLength2': 'value 必须是长度为2的数组: ',
        'notArrayConsistNumbers': 'value 数组必须是数字: ',
        'invalidValueRange': 'value 区间无效（左 >= 右）: ',
        'unsupportedValue': '不支持的 value 字符串值: ',
        'valueTypeError': 'value 必须是数字数组或字符串: ',
        'descriptionError': 'description 包含恶意内容或格式错误: ',
        'circularReferenceError1': '检测到循环引用：节点 ',
        'circularReferenceError2': ' 已存在于路径中',
        'invalidTag': '无效标签: ',
        'pathMayNotInSecureDirectory': '该路径可能不在安全目录内',
        'serverError': '服务异常，请稍后重试'
    },
    EN: {
        'typeError': 'File type error',
        'searchTypeError': 'Search type error',
        'nodeInfoError': 'The query parameter nodeInfo and its sub-items must exist and must be a dictionary.',
        'metaDataError': 'The query parameter “metaData” and its sub-items must exist and must be a dictionary.',
        'dbInitError': 'Database initialization failed',
        'loadConfigError': 'load graph config info failed',
        'rankStepNullError': 'rank or step or micro_step is null',
        'loadNodeError': 'load graph all node list failed',
        'expandNodeError': 'node expand or collapse error',
        'searchNodeError': 'node search error',
        'graphTypeError': 'graph type error, only support "NPU" or "BENCH"',
        'getNodeInfoError': 'get node info failed',
        'taskTypeError': 'task type not support',
        'matchNodeError': 'match node failed',
        'loadConfigFileError': 'load config file failed',
        'parameterError': 'Parameter Error',
        'nodesTypeError': 'Node types are inconsistent; matching relationships cannot be added.',
        'taskTypeWrong': 'Task type error',
        'nodeNotExist': 'Node does not exist',
        'ioEmptyError': 'The input or output statistical error value is empty.',
        'outputEmptyError': 'The output statistical error value is empty, and the calculation precision error failed.',
        'operationFailed': 'Operation failed: ',
        'or': 'or',
        'nodeUnmatchError': 'Node not matched. Please match the node first.',
        'deleteMatchNodeFailed': 'Failed to delete matching node',
        'noUpdateNodesFound': 'No nodes found to update.',
        'updateNodeAccuracyFailed': 'Failed to update node accuracy',
        'updateDatabaseFailed': 'Failed to update the database',
        'updateColorFailed': 'Color update failed',
        'saveMatchRelationshipFailed': 'Failed to save matching relationship',
        'selectedNodeCantMatched': 'The selected node cannot be matched.',
        'loadErrorBySecurityIssue': 'The file has security issues, and reading the file failed.',
        'loadConfigFailed': 'Failed to retrieve configuration information. Please check the first file in directory.',
        'debug': 'Debug',
        'bench': 'Bench',
        'expandOrCollapseError': 'An error occurred while expanding or collapsing the node.',
        'obtainNodesAccuracyFailed': 'Failed to obtain nodes meeting accuracy error requirements',
        'notSupportOverflow': 'Overflow detection is not supported in multi-image mode.',
        'obtainNodesOverflowFailed': 'Failed to obtain overflow detection node',
        'updatePrecisionFailed': 'Update precision error failure',
        'nodeTypeError': 'Node type error',
        'readFileFailed': 'Failed to read file',
        'parameterEmpty': 'The parameter is empty.',
        'saveFileFailed': 'Failed to save file',
        'dbConncetionNotInit': 'database connection not init',
        'parseJsonFailed': 'Failed to parse JSON',
        'getNodeListFailed': 'Failed to get node list',
        'initGraphJsonFailed': 'initlize graph json failed',
        'colorsNotObject': 'Colors must be an object',
        'colotsEmpty': 'Colors cannot be empty',
        'illegalColorKey': 'Illegal color key: ',
        'colorValuesNotObject': 'Color values must be objects: ',
        'missingValueField': 'Missing value field: ',
        'notArrayOfLength2': 'The value must be an array of length 2: ',
        'notArrayConsistNumbers': 'Value array must consist of numbers: ',
        'invalidValueRange': 'Invalid value range (left >= right): ',
        'unsupportedValue': 'Unsupported value string values: ',
        'valueTypeError': 'Value must be a numeric array or string: ',
        'descriptionError': 'Description contains malicious content or formatting errors: ',
        'circularReferenceError1': 'Detected circular reference: node ',
        'circularReferenceError2': ' already exists in the path',
        'invalidTag': 'Invalid tag: ',
        'pathMayNotInSecureDirectory': 'The path may not be within a secure directory',
        'serverError': 'server error, please try again later'
    }
}



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from collections import defaultdict

from msprobe.core.common.const import CompareConst, Const
from msprobe.core.common.file_utils import load_json, load_yaml, save_yaml
from msprobe.core.common.utils import add_time_with_yaml, detect_framework_by_dump_json, \
    get_stack_construct_by_dump_json_path, CompareException
from msprobe.core.compare.layer_mapping.data_scope_parser import get_dump_data_items
from msprobe.core.compare.utils import read_op, reorder_op_name_list
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.core.common.log import logger


class LayerTrie:
    def __init__(self, type_name, framework=None):
        self.type_name = type_name
        self.data_items = defaultdict(list)
        self.children = {}
        self.framework = framework

    def __repr__(self):
        data_nums = [{k: len(v)} for k, v in self.data_items.items()]
        return f"Layer(type_name={self.type_name}, data_number={data_nums})"

    def get(self, name):
        return self.children.get(name)

    def insert(self, data_item):
        parts = data_item.full_scope.split(Const.SEP)
        node = self
        scope_name_list = parts[Const.RIGHT_MOVE_INDEX:]

        for name in scope_name_list:
            if name not in node.children:
                node.children[name] = LayerTrie(name, data_item.framework)
            node = node.children[name]
        node.data_items[data_item.state].append(data_item)
        node.type_name = data_item.type_name

    def query_data(self, scope, state, index, default_value=None):
        parts = scope.split(Const.SEP)
        node = self
        scope_name_list = parts[1:]

        for name in scope_name_list:
            if name not in node.children:
                return default_value
            node = node.children[name]
        if index >= len(node.data_items[state]):
            return default_value
        if node.data_items[state]:
            return node.data_items[state][index]
        else:
            logger.error(f"node.data_items of state:{state} is empty, please check.")
            raise CompareException(CompareException.INDEX_OUT_OF_BOUNDS_ERROR)

    def save_to_yaml(self, output_path):
        result = {f"{self.type_name} @ {self}": self.convert_to_dict(self)}
        file_name = add_time_with_yaml(f"{self.framework}_tree")
        file_path = os.path.join(os.path.realpath(output_path), file_name)
        save_yaml(file_path, result)

    @recursion_depth_decorator("LayerMapping: LayerTrie.convert_to_dict", max_depth=100)
    def convert_to_dict(self, node):
        result = {}
        result["data_item"] = {st: [dt.data_name for dt in dts] for st, dts in node.data_items.items()}
        for child_key, child_node in node.children.items():
            key = f"{child_key} @ {child_node}"
            result[key] = self.convert_to_dict(child_node)
        return result


def convert_scope(layer_trie, data_item, mapping=None):
    if not mapping:
        mapping = {}
    new_scope = Const.TOP_LAYER
    scope_list = data_item.full_scope.split(Const.SEP)
    cur_node = layer_trie

    idx = 0
    while idx < len(scope_list) - 1:
        child_name = scope_list[idx + 1]
        type_name = cur_node.type_name
        prefix_mapping = mapping.get(type_name, {})
        mapping_list = prefix_mapping.get(child_name, [])
        mapping_list.append((child_name, child_name, 1))
        step = 1
        for origin, target, level in mapping_list:
            if Const.SEP.join(scope_list[idx + 1: idx + level + 1]) == origin:
                new_scope = new_scope + Const.SEP + target
                step = level
                break
        for _ in range(step):
            child_node = cur_node.get(scope_list[idx + 1])
            cur_node = child_node
            idx += 1
    index = -1
    state = data_item.state
    for idx, child in enumerate(cur_node.data_items[state]):
        if data_item.data_name == child.data_name:
            index = idx
    return new_scope, state, index


def get_data_items_and_tree(dump_json_path, output_path):
    framework = detect_framework_by_dump_json(dump_json_path)
    stack, construct = get_stack_construct_by_dump_json_path(dump_json_path)
    dump = load_json(dump_json_path)
    dump_data_items = get_dump_data_items(dump, stack, construct, framework, output_path)
    root = LayerTrie(Const.TOP_LAYER, framework)
    for data_item in dump_data_items:
        root.insert(data_item)
    if output_path:
        root.save_to_yaml(output_path)
    return dump_data_items, root


def convert_data_item(npu_tree, bench_tree, npu_data_item, mapping):
    new_scope, state, index = convert_scope(npu_tree, npu_data_item, mapping)
    bench_data_item = bench_tree.query_data(new_scope, state, index)
    return bench_data_item


def update_keys_in_place(d):
    """
    This function is used to compare and maintain compatibility between the old and new versions.
    In the old version, 'Cell' was used as the top layer name, while the new version uses 'TopLayer'.
    """
    cell_value = d.pop(Const.CELL, None)

    if cell_value is not None:
        d[Const.TOP_LAYER] = cell_value


def preprocess_layer_mapping(mapping):
    """
    before:
        {'A': {'a.b.c': 'new_c',
               'a.demo': 'new_demo',
               'z': 'new_z',
               'd.e': 'e'}}
    after:
        {'A': {'a': [('a.b.c', 'new_c', 3), ('a.demo', 'new_demo', 2)],
               'z': [('z', 'new_z', 1)],
               'd': [('d.e', 'e', 2)]}}
    """
    update_keys_in_place(mapping)
    final_mapping = {}

    for type_name, name_map in mapping.items():
        final_mapping[type_name] = {}

        for key, value in name_map.items():
            key_list = key.split('.')
            prefix = key_list[0]  # 取前缀
            value_list = value.split('(')
            value = value_list[0]  # 取前缀
            key_len = len(key_list)
            if prefix not in final_mapping[type_name]:
                final_mapping[type_name][prefix] = []
            final_mapping[type_name][prefix].append((key, value, key_len))

        # 前缀映射列表按规则长度排序
        for prefix in final_mapping[type_name]:
            final_mapping[type_name][prefix].sort(key=lambda x: -x[-1])

    return final_mapping


def convert_data_items(npu_tree, bench_tree, npu_data_items, mapping):
    mapping = preprocess_layer_mapping(mapping)
    api_mapping = {}
    for npu_data_item in npu_data_items:
        bench_data_item = convert_data_item(npu_tree, bench_tree, npu_data_item, mapping)
        bench_name = bench_data_item.data_name if bench_data_item else CompareConst.N_A
        npu_name = npu_data_item.data_name
        api_mapping[npu_name] = bench_name
    return api_mapping


def generate_api_mapping_by_layer_mapping(npu_json_path, bench_json_path, layer_mapping_path=None, output_path=None):
    npu_data_items, npu_root = get_data_items_and_tree(npu_json_path, output_path)
    _, bench_root = get_data_items_and_tree(bench_json_path, output_path)
    if isinstance(layer_mapping_path, str):
        mapping = load_yaml(layer_mapping_path)
    else:
        mapping = {}
    api_mapping = convert_data_items(npu_root, bench_root, npu_data_items, mapping)
    if output_path:
        file_name = add_time_with_yaml("api_mapping")
        file_path = os.path.join(os.path.realpath(output_path), file_name)
        save_yaml(file_path, api_mapping)
    return api_mapping


def generate_data_mapping(npu_json_path, bench_json_path, api_mapping, output_path=None):
    def read_full_op_names(data, op_name):
        op_parsed_list = read_op(data.get(op_name, {}), op_name)
        full_op_names = [op_parsed.get('full_op_name') for op_parsed in op_parsed_list]
        states = [op_parsed.get(Const.STATE) for op_parsed in op_parsed_list]
        return full_op_names, states

    def generate_op_data_mapping(npu_op_name, npu_full_op_names, bench_op_name, bench_full_op_names):
        suffix_to_full_op_name = {}
        op_data_mapping = {}
        for bench_full_op_name in bench_full_op_names:
            suffix = bench_full_op_name[len(bench_op_name):]
            suffix_to_full_op_name[suffix] = bench_full_op_name

        for npu_full_op_name in npu_full_op_names:
            suffix = npu_full_op_name[len(npu_op_name):]
            op_data_mapping[npu_full_op_name] = suffix_to_full_op_name.get(suffix, CompareConst.N_A)
        return op_data_mapping

    npu_data = load_json(npu_json_path).get("data", {})
    bench_data = load_json(bench_json_path).get("data", {})
    data_mapping = {}
    for npu_op_name, bench_op_name in api_mapping.items():
        if not npu_op_name:
            continue
        npu_full_op_names, npu_states = read_full_op_names(npu_data, npu_op_name)
        bench_full_op_names, bench_states = read_full_op_names(bench_data, bench_op_name)
        npu_full_op_names_reorder, _ = reorder_op_name_list(npu_full_op_names, npu_states)
        bench_full_op_names_reorder, _ = reorder_op_name_list(bench_full_op_names, bench_states)
        mapping = generate_op_data_mapping(npu_op_name, npu_full_op_names_reorder,
                                           bench_op_name, bench_full_op_names_reorder)
        data_mapping.update(mapping)
    if output_path:
        file_name = add_time_with_yaml("data_mapping")
        file_path = os.path.join(os.path.realpath(output_path), file_name)
        save_yaml(file_path, data_mapping)
    return data_mapping


def generate_data_mapping_by_layer_mapping(input_param, layer_mapping_path=None, output_path=None):
    npu_json_path = input_param.get("npu_path")
    bench_json_path = input_param.get("bench_path")
    api_mapping = generate_api_mapping_by_layer_mapping(
        npu_json_path, bench_json_path, layer_mapping_path)
    data_mapping = generate_data_mapping(
        npu_json_path, bench_json_path, api_mapping, output_path)
    return data_mapping



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class mainly involves generate npu dump data function.
"""
import json
import sys
import os
import stat
import re
import shutil

import numpy as np

from msprobe.core.common.log import logger
from msprobe.core.common.file_utils import check_file_or_directory_path, create_directory, save_npy
from msprobe.infer.offline.compare.msquickcmp.atc import atc_utils
from msprobe.infer.offline.compare.msquickcmp.common import utils
from msprobe.infer.offline.compare.msquickcmp.common.dump_data import DumpData
from msprobe.infer.offline.compare.msquickcmp.common.utils import AccuracyCompareException, parse_input_shape_to_list
from msprobe.infer.offline.compare.msquickcmp.common.dynamic_argument_bean import DynamicArgumentEnum
from msprobe.infer.offline.compare.msquickcmp.npu.om_parser import OmParser
from msprobe.infer.utils.check.rule import Rule
from msprobe.infer.utils.util import load_file_to_read_common_check, filter_cmd
from msprobe.infer.utils.file_open_check import ms_open


BENCHMARK_DIR = "benchmark"
ACL_JSON_PATH = "acl.json"
NPU_DUMP_DATA_BASE_PATH = "dump_data/npu"
NPU_DUMP_DATA_GOLDEN_PATH = "dump_data/npu_golden"
RESULT_DIR = "result"
INPUT = "input"
INPUT_SHAPE = "--input_shape"
OUTPUT_SIZE = "--outputSize"
INPUT_FORMAT_TO_RGB_RATIO_DICT = {"YUV420SP_U8": 2, "RGB888_U8": 1, "YUV400_U8": 3, "XRGB8888_U8": 3 / 4}
OPEN_FLAGS = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
OPEN_MODES = stat.S_IWUSR | stat.S_IRUSR
DTYPE_MAP = {
    "dtype.float32": np.float32,
    "dtype.float16": np.float16,
    "dtype.float64": np.float64,
    "dtype.int8": np.int8,
    "dtype.int16": np.int16,
    "dtype.int32": np.int32,
    "dtype.int64": np.int64,
    "dtype.uint8": np.uint8,
    "dtype.uint16": np.uint16,
    "dtype.uint32": np.uint32,
    "dtype.uint64": np.uint64,
    "dtype.bool": np.bool_,
}


class DynamicInput(object):
    def __init__(self, om_parser, input_shape):
        self.input_shape = input_shape
        self.om_parser = om_parser
        self.atc_dynamic_arg, self.cur_dynamic_arg = self.get_dynamic_arg_from_om(om_parser)
        self.dynamic_arg_value = self.get_arg_value(om_parser, input_shape)

    @staticmethod
    def get_dynamic_arg_from_om(om_parser):
        atc_cmd_args = om_parser.get_atc_cmdline().split(" ")
        for i, atc_arg in enumerate(atc_cmd_args):
            for dym_arg in DynamicArgumentEnum:
                if dym_arg.value.atc_arg in atc_arg:
                    if dym_arg.value.atc_arg == atc_arg:
                        atc_arg += '=' + atc_cmd_args[i + 1]
                    return atc_arg, dym_arg
        return "", None

    @staticmethod
    def get_input_shape_from_om(om_parser):
        # get atc input shape from atc cmdline
        atc_input_shape = ""
        atc_cmd_args = om_parser.get_atc_cmdline().split(" ")
        for i, atc_arg in enumerate(atc_cmd_args):
            if INPUT_SHAPE == atc_arg:
                atc_input_shape = atc_cmd_args[i + 1]
                break
            if atc_arg.startswith(INPUT_SHAPE + utils.EQUAL):
                atc_input_shape = atc_arg.split(utils.EQUAL)[1]
                break
        return atc_input_shape

    @staticmethod
    def get_arg_value(om_parser, input_shape):
        is_dynamic_scenario, scenario = om_parser.get_dynamic_scenario_info()
        if not is_dynamic_scenario:
            logger.info("The input of model is not dynamic.")
            return ""
        if om_parser.shape_range or scenario == DynamicArgumentEnum.DYM_DIMS:
            return input_shape
        atc_input_shape = DynamicInput.get_input_shape_from_om(om_parser)
        # from atc input shape and current input shape to get input batch size
        # if dim in shape is -1, the shape in the index of current input shape is the batch size
        atc_input_shape_dict = utils.parse_input_shape(atc_input_shape)
        quickcmp_input_shape_dict = utils.parse_input_shape(input_shape)
        batch_size_set = set()
        for op_name in atc_input_shape_dict.keys():
            DynamicInput.get_dynamic_dim_values(
                atc_input_shape_dict.get(op_name), quickcmp_input_shape_dict.get(op_name), batch_size_set
            )
        if len(batch_size_set) == 1:
            for batch_size in batch_size_set:
                return str(batch_size)
        logger.error("Please check your input_shape arg is valid.")
        raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR)

    @staticmethod
    def get_dynamic_dim_values(dym_shape, cur_shape, shape_values):
        for dim, _ in enumerate(dym_shape):
            if dym_shape[dim] != "-1":
                continue
            if isinstance(shape_values, list):
                shape_values.append(int(cur_shape[dim]))
            else:
                shape_values.add(cur_shape[dim])

    def add_dynamic_arg_for_benchmark(self, benchmark_cmd: list):
        if self.is_dynamic_shape_scenario():
            self.check_input_dynamic_arg_valid()
            benchmark_cmd.append(self.cur_dynamic_arg.value.benchmark_arg)
            benchmark_cmd.append(self.dynamic_arg_value)

    def is_dynamic_shape_scenario(self):
        """
        if atc cmdline contain dynamic argument
        """
        return self.atc_dynamic_arg != ""

    def judge_dynamic_shape_scenario(self, atc_dym_arg):
        """
        check the dynamic shape scenario
        """
        return self.atc_dynamic_arg.split(utils.EQUAL)[0] == atc_dym_arg

    def check_input_dynamic_arg_valid(self):
        if self.cur_dynamic_arg is DynamicArgumentEnum.DYM_SHAPE:
            return
        # check dynamic input value is valid, "--arg=value" ,split by '='
        dynamic_arg_values = self.atc_dynamic_arg.split(utils.EQUAL)[1]
        if self.judge_dynamic_shape_scenario(DynamicArgumentEnum.DYM_DIMS.value.atc_arg):
            self.check_dynamic_dims_valid(dynamic_arg_values)
            return
        if self.judge_dynamic_shape_scenario(DynamicArgumentEnum.DYM_BATCH.value.atc_arg):
            self.check_dynamic_batch_valid(dynamic_arg_values)

    def check_dynamic_batch_valid(self, atc_dynamic_arg_values):
        dynamic_arg_values = atc_dynamic_arg_values.replace(utils.COMMA, utils.SEMICOLON)
        try:
            atc_value_list = utils.parse_arg_value(dynamic_arg_values)
            cur_input = utils.parse_value_by_comma(self.dynamic_arg_value)
        except AccuracyCompareException as err:
            logger.error(f"Please input the valid shape, ' 'the valid dynamic value range are {dynamic_arg_values}")
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR) from err
        for value in atc_value_list:
            if cur_input == value:
                return

    def check_dynamic_dims_valid(self, atc_dynamic_arg_values):
        atc_input_shape = DynamicInput.get_input_shape_from_om(self.om_parser)
        try:
            atc_input_shape_dict = utils.parse_input_shape(atc_input_shape)
            quickcmp_input_shape_dict = utils.parse_input_shape(self.dynamic_arg_value)
            dym_dims = []
            for op_name in atc_input_shape_dict.keys():
                DynamicInput.get_dynamic_dim_values(
                    atc_input_shape_dict.get(op_name), quickcmp_input_shape_dict.get(op_name), dym_dims
                )
            atc_value_list = utils.parse_arg_value(atc_dynamic_arg_values)
        except AccuracyCompareException as err:
            logger.error(f"Please input the valid shape, ' 'the valid dynamic value range are {atc_dynamic_arg_values}")
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR) from err
        for value in atc_value_list:
            if dym_dims == value:
                return


class NpuDumpData(DumpData):
    """
    Class for generate npu dump data
    """

    def __init__(self, arguments, is_golden=False):
        super().__init__()
        self.target_path = arguments.golden_path if is_golden else arguments.target_path
        self.output_path = arguments.output_path
        self.input_data = arguments.input_data
        self.input_shape = arguments.input_shape
        self.output_size = arguments.output_size
        self.device, self.is_golden = arguments.rank, is_golden
        self.dump = True

        self.benchmark_input_path = ""
        output_json_path = atc_utils.convert_model_to_json(arguments.cann_path, self.target_path, self.output_path)
        self.om_parser = OmParser(output_json_path)
        self.dynamic_input = DynamicInput(self.om_parser, self.input_shape)
        self.python_version = sys.executable or "python3"
        self.data_dir = self._create_dir()

    @staticmethod
    def _write_content_to_acl_json(acl_json_path, model_name, npu_data_output_dir, sub_model_name_list=None):
        load_dict = {
            "dump": {
                "dump_list": [{"model_name": model_name}],
                "dump_path": npu_data_output_dir,
                "dump_mode": "all",
                "dump_op_switch": "off",
            }
        }
        load_dict["dump"]["dump_list"].extend([{"model_name": ii} for ii in sub_model_name_list])

        if os.access(acl_json_path, os.W_OK):
            json_stat = os.stat(acl_json_path)
            if json_stat.st_uid == os.getuid():
                os.remove(acl_json_path)
            else:
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_PARSER_JSON_FILE_ERROR)
            try:
                with ms_open(acl_json_path, "w") as write_json:
                    try:
                        json.dump(load_dict, write_json)
                    except ValueError as exc:
                        logger.error(str(exc))
                        raise AccuracyCompareException(utils.ACCURACY_COMPARISON_WRITE_JSON_FILE_ERROR) from exc
            except IOError as acl_json_file_except:
                logger.error('Failed to open"' + acl_json_path + '", ' + str(acl_json_file_except))
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_OPEN_FILE_ERROR) from acl_json_file_except
        else:
            logger.error(f"The path {acl_json_path} does not have permission to write. "
                         f"Please check the path permission")
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PATH_ERROR)

    def generate_inputs_data(self, npu_dump_data_path=None, use_aipp=False):
        if self.input_data:
            input_path = self.input_data.split(",")
            for i, input_file in enumerate(input_path):
                if not os.path.isfile(input_file):
                    logger.error(f"no such file exists: {input_file}")
                    raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
                file_name = "input_" + str(i) + ".bin"
                dest_file = os.path.join(self.output_path, "input", file_name)
                shutil.copy(input_file, dest_file)
                os.chmod(dest_file, 0o640)
            return
        if use_aipp:
            self._generate_inputs_data_for_aipp(self.data_dir)
        else:
            self._generate_inputs_data_without_aipp(self.data_dir)

    def generate_dump_data(self, npu_dump_path=None, om_parser=None):
        """
        Function Description:
            compile and rum benchmark project
        Return Value:
            npu dump data path
        """
        self._check_input_path_param()
        return self.benchmark_run()

    def get_expect_output_name(self):
        """
        Function Description:
            get expect output node name in golden net
        Return Value:
            output node name in golden net
        """
        return self.om_parser.get_expect_net_output_name()

    def benchmark_run(self):
        """
        Function Description:
            run benchmark project
        Return Value:
            npu dump data path
        Exception Description:
            when invalid npu dump data path throw exception
        """
        try:
            import ais_bench
        except ModuleNotFoundError as err:
            raise err

        self._compare_shape_vs_file()
        npu_data_output_dir = os.path.join(
            self.output_path,
            NPU_DUMP_DATA_GOLDEN_PATH if self.is_golden else NPU_DUMP_DATA_BASE_PATH
        )
        create_directory(npu_data_output_dir)
        model_name, extension = utils.get_model_name_and_extension(self.target_path)
        acl_json_path = os.path.join(npu_data_output_dir, ACL_JSON_PATH)
        if not os.path.exists(acl_json_path):
            os.mknod(acl_json_path, mode=0o600)
        benchmark_cmd = [
            self.python_version,
            "-m",
            "ais_bench",
            "--model",
            self.target_path,
            "--input",
            self.benchmark_input_path,
            "--device",
            self.device,
            "--output",
            npu_data_output_dir,
            "--warmup_count",
            "0",
        ]
        if self.dump:
            cur_dir = os.getcwd()
            acl_json_path = os.path.join(cur_dir, acl_json_path)
            sub_model_name_list = self.om_parser.get_sub_graph_name()
            self._write_content_to_acl_json(acl_json_path, model_name, npu_data_output_dir, sub_model_name_list)
            benchmark_cmd.extend(["--acl_json_path", acl_json_path])

        self.dynamic_input.add_dynamic_arg_for_benchmark(benchmark_cmd)
        if self.output_size or self.dynamic_input.is_dynamic_shape_scenario():
            self._make_benchmark_cmd_for_shape_range(benchmark_cmd)

        # do benchmark command
        benchmark_cmd = filter_cmd(benchmark_cmd)
        utils.execute_command(benchmark_cmd, False)

        npu_dump_data_path = ""
        if self.dump:
            npu_dump_data_path, file_is_exist = utils.get_dump_data_path(npu_data_output_dir, False, model_name)
            if not file_is_exist:
                logger.error(f"The path {npu_dump_data_path} dump data is not exist.")
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PATH_ERROR)

        # net output data path
        npu_net_output_data_path, file_is_exist = utils.get_dump_data_path(npu_data_output_dir, True, model_name)
        if not file_is_exist:
            logger.error(f"The path {npu_net_output_data_path} net output data is not exist.")
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PATH_ERROR)

        self._convert_net_output_to_numpy(npu_net_output_data_path, npu_dump_data_path)
        if self.is_golden:
            return npu_dump_data_path, ""
        else:
            return npu_dump_data_path, npu_net_output_data_path

    def _create_dir(self):
        data_dir = os.path.join(self.output_path, "input")
        create_directory(data_dir)
        return data_dir

    def _get_inputs_info_from_aclruntime(self):
        import aclruntime

        options = aclruntime.session_options()
        Rule.input_file().check(self.target_path, will_raise=True)
        self.target_path = load_file_to_read_common_check(self.target_path)
        aa = aclruntime.InferenceSession(self.target_path, int(self.device), options)
        shape_list = [ii.shape for ii in aa.get_inputs()]
        dtype_list = [ii.datatype.name for ii in aa.get_inputs()]

        aa.free_resource()
        return shape_list, dtype_list

    def _generate_inputs_data_without_aipp(self, input_dir):
        if os.listdir(input_dir):
            return

        inputs_list, data_type_list = self._get_inputs_info_from_aclruntime()
        if self.dynamic_input.is_dynamic_shape_scenario() and not self.input_shape:
            logger.error("Please set '--input_shape' to fix the dynamic shape.")
            raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR)

        if self.input_shape:
            inputs_list = parse_input_shape_to_list(self.input_shape)

        for i, (input_shape, data_type) in enumerate(zip(inputs_list, data_type_list)):
            input_data = np.random.random(input_shape).astype(data_type)
            file_name = "input_" + str(i) + ".bin"
            input_data.tofile(os.path.join(input_dir, file_name))
            os.chmod(os.path.join(input_dir, file_name), 0o640)

    def _generate_inputs_data_for_aipp(self, input_dir):
        aipp_contents = self.om_parser.get_aipp_config_content()
        src_image_size_h = []
        src_image_size_w = []
        input_format = []
        for aipp_content in aipp_contents:
            aipp_list = aipp_content.split(",")
            for aipp_info in aipp_list:
                aipp_infos_split_by_colon = aipp_info.split(":")
                if len(aipp_infos_split_by_colon) < 2:
                    continue
                if "src_image_size_h" in aipp_info:
                    src_image_size_h.append(aipp_infos_split_by_colon[1])
                if "src_image_size_w" in aipp_info:
                    src_image_size_w.append(aipp_infos_split_by_colon[1])
                if "input_format" in aipp_info:
                    input_format.append(aipp_infos_split_by_colon[1].strip('\\"'))
            if not src_image_size_h or not src_image_size_w:
                logger.error("atc insert_op_config file contains no src_image_size_h or src_image_size_w")
                raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_WRONG_AIPP_CONTENT)
            if len(src_image_size_h) != len(src_image_size_w):
                logger.error("atc insert_op_config file's src_image_size_h number does not equal src_image_size_w")
                raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_WRONG_AIPP_CONTENT)
        if self.input_shape:
            inputs_list = parse_input_shape_to_list(self.input_shape)
        else:
            inputs_list = self.om_parser.get_shape_list()
            if len(inputs_list) != len(src_image_size_h):
                logger.error("inputs number is not equal to aipp inputs number, please check the -is param")
                raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_WRONG_AIPP_CONTENT)
        # currently, onnx only support input format nchw
        h_position = 2
        w_position = 3
        input_dir = os.path.join(self.output_path, "input")
        for i, item in enumerate(inputs_list):
            item[h_position] = int(src_image_size_h[i])
            item[w_position] = int(src_image_size_w[i])
            div_input_format = INPUT_FORMAT_TO_RGB_RATIO_DICT.get(input_format[i])
            if not div_input_format:
                logger.error("aipp input format only support: YUV420SP_U8, RGB888_U8, YUV400_U8, XRGB8888_U8")
                raise utils.AccuracyCompareException(utils.ACCURACY_COMPARISON_WRONG_AIPP_CONTENT)
            input_data = np.random.randint(0, 256, int(np.prod(item) / div_input_format)).astype(np.uint8)
            file_name = "input_" + str(i) + ".bin"
            input_data.tofile(os.path.join(input_dir, file_name))
            os.chmod(os.path.join(input_dir, file_name), 0o640)

    def _make_benchmark_cmd_for_shape_range(self, benchmark_cmd):
        pattern = re.compile(r'^[0-9]+$')
        count = self.om_parser.get_net_output_count()
        if not self.output_size:
            if count > 0:
                count_list = []
                for _ in range(count):
                    count_list.append("90000000")
                self.output_size = ",".join(count_list)
        if self.output_size:
            output_size_list = self.output_size.split(',')
            if len(output_size_list) != count:
                logger.error(
                    f"The output size ({len(output_size_list)}) is not equal {count} in model."
                    f" Please check the '--output-size' argument.")
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
            for item in output_size_list:
                item = item.strip()
                match = pattern.match(item)
                if match is None:
                    logger.error(f"The size ({self.output_size}) is invalid. Please check the output size.")
                    raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
                if int(item) <= 0:
                    logger.error(f"The size ({self.output_size}) must be large than zero. "
                                 f"Please check the output size.")
                    raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
            benchmark_cmd.append(OUTPUT_SIZE)
            benchmark_cmd.append(self.output_size)

    def _convert_net_output_to_numpy(self, npu_net_output_data_path, npu_dump_data_path):
        net_output_data = None
        npu_net_output_data_info = self.om_parser.get_net_output_data_info(npu_dump_data_path)
        for dir_path, _, files in os.walk(npu_net_output_data_path):
            for index, each_file in enumerate(sorted(files)):
                data_type = npu_net_output_data_info.get(index)[0]
                shape = npu_net_output_data_info.get(index)[1]
                data_len = utils.get_data_len_by_shape(shape)
                each_file_path = os.path.join(dir_path, each_file)
                if Rule.input_file().check(each_file_path, will_raise=True):
                    original_net_output_data = np.fromfile(each_file_path, data_type, data_len)
                try:
                    net_output_data = original_net_output_data.reshape(shape)
                except ValueError:
                    logger.warning(f"The shape of net_output data from file {each_file} is {shape}.")
                    net_output_data = original_net_output_data
                    
                each_file_index = each_file.split('_')[-1]
                new_each_file = "output_" + each_file_index
                file_name = os.path.basename(new_each_file).split('.')[0]
                numpy_file_path = os.path.join(npu_net_output_data_path, file_name)
                save_npy(net_output_data, numpy_file_path)
                new_each_file_path = os.path.join(dir_path, new_each_file)
                os.rename(each_file_path, new_each_file_path)

    def _check_input_path_param(self):
        if self.input_data == "":
            input_path = os.path.join(self.output_path, INPUT)
            check_file_or_directory_path(os.path.realpath(input_path), True)
            input_bin_files = os.listdir(input_path)
            input_bin_files.sort(key=lambda file: int((re.findall("\\d+", file))[0]))
            bin_file_path_array = []
            for item in input_bin_files:
                bin_file_path_array.append(os.path.join(input_path, item))
            self.benchmark_input_path = ",".join(bin_file_path_array)
        else:
            bin_file_path_array = utils.check_input_bin_file_path(self.input_data)
            self.benchmark_input_path = ",".join(bin_file_path_array)

    def _compare_shape_vs_file(self):
        shape_size_array = self.om_parser.get_shape_size()
        if self.om_parser.contain_negative_1:
            return
        files_size_array = self._get_file_size()
        if not self.om_parser.get_aipp_config_content:
            self._shape_size_vs_file_size(shape_size_array, files_size_array)

    def _get_file_size(self):
        file_size = []
        files = self.benchmark_input_path.split(",")
        for item in files:
            if item.endswith("bin") or item.endswith("BIN"):
                file_size.append(os.path.getsize(item))
            elif item.endswith("npy") or item.endswith("NPY"):
                try:
                    Rule.input_file().check(item, will_raise=True)
                    file_size.append(np.load(item).size)
                except (ValueError, FileNotFoundError) as e:
                    logger.error(f"The path {item} can not get its size through numpy")
                    raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PATH_ERROR) from e
            else:
                logger.error(f"Input_path parameter only support bin or npy file, ' 'but got {item}")
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PATH_ERROR)
        return file_size

    def _shape_size_vs_file_size(self, shape_size_array, files_size_array):
        if len(shape_size_array) < len(files_size_array):
            logger.error("The number of input files is incorrect.")
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_BIN_FILE_ERROR)
        if self.om_parser.shape_range:
            for file_size in files_size_array:
                if file_size not in shape_size_array:
                    logger.error(f"The size ({file_size}) of file can not match the input of the model.")
                    raise AccuracyCompareException(utils.ACCURACY_COMPARISON_BIN_FILE_ERROR)
        elif self.dynamic_input.is_dynamic_shape_scenario():
            for shape_size in shape_size_array:
                for file_size in files_size_array:
                    if file_size <= shape_size:
                        return
            logger.warning("The size of bin file can not match the input of the model.")
        else:
            for shape_size, file_size in zip(shape_size_array, files_size_array):
                if shape_size == 0:
                    continue
                if shape_size != file_size:
                    logger.error("The shape value is different from the size of the file.")
                    raise AccuracyCompareException(utils.ACCURACY_COMPARISON_BIN_FILE_ERROR)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.log import logger
from msprobe.infer.offline.compare.msquickcmp.main import install_offline_deps_cli


def _install_deps_parser(parser):
    parser.add_argument("-m", "--mode", dest="mode", type=str,
                        help="Install deps mode: 'offline' for offline_dump om and compare offline_model",
                        required=True)
    parser.add_argument("--no_check", dest="no_check", action="store_true",
                        help="<optional> Whether to skip checking the target website's certificate information "
                             "when installing dependency packages poses a certain security risk. "
                             "This poses a certain security risk, "
                             "and users should use it with caution and bear the consequences themselves.",
                        required=False)


MODE_DISPATCHER = {
    'offline': install_offline_deps_cli
}


def install_deps_cli(args):
    """
    Dispatch install deps based on mode parameter
    """
    mode = getattr(args, "mode", None)
    if not mode:
        logger.error("No mode specified, please check.")
    # Get the appropriate function based on mode
    install_deps_func = MODE_DISPATCHER.get(mode)
    if install_deps_func is None:
        logger.error(f"Invalid install deps mode: '{mode}'. Available modes: {list(MODE_DISPATCHER.keys())}")
        raise RuntimeError
    # Execute the install_deps function
    return install_deps_func(args)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from mindspore import nn

from msprobe.core.common.const import Const
from msprobe.core.common.exceptions import MsprobeException
from msprobe.core.common.file_utils import create_directory
from msprobe.core.common.log import logger
from msprobe.mindspore.common.const import Const as MsConst
from msprobe.mindspore.common.utils import is_mindtorch

if is_mindtorch():
    import torch


class DebuggerConfig:
    def __init__(self, common_config, task_config):
        self.execution_mode = None
        self.dump_path = common_config.dump_path
        self.task = common_config.task
        self.rank = [] if not common_config.rank else common_config.rank
        self.step = [] if not common_config.step else common_config.step
        common_config.level = Const.LEVEL_L1 if not common_config.level else common_config.level
        self.level = MsConst.TOOL_LEVEL_DICT.get(common_config.level, MsConst.API)
        self.level_ori = common_config.level
        self.list = [] if not task_config.list else task_config.list
        self.scope = [] if not task_config.scope else task_config.scope
        self.data_mode = [Const.ALL] if not task_config.data_mode else task_config.data_mode
        self.file_format = task_config.file_format
        self.check_mode = task_config.check_mode
        self.framework = Const.MS_FRAMEWORK
        self.summary_mode = task_config.summary_mode
        self.stat_cal_mode = task_config.stat_cal_mode if hasattr(task_config, 'stat_cal_mode') else None
        self.device_stat_precision_mode = task_config.device_stat_precision_mode \
            if hasattr(task_config, 'device_stat_precision_mode') else None
        self.async_dump = common_config.async_dump if common_config.async_dump else False
        self.precision = common_config.precision if common_config.precision else Const.DUMP_PRECISION_LOW
        self.check()
        self._check_statistics_config(task_config)
        create_directory(self.dump_path)

    @staticmethod
    def check_model(models, token_range=None):
        if token_range and not models:
            error_info = "The 'model' parameter must be provided when token_range is not None"
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR, error_info)

        target_module_type = (torch.nn.Module, "torch.nn.Module") if is_mindtorch() else (nn.Cell, "mindspore.nn.Cell")
        if models is None or isinstance(models, target_module_type[0]):
            return models
        if isinstance(models, (list, tuple)):
            error_model = None
            for model in models:
                if not isinstance(model, target_module_type[0]):
                    error_model = model
                    break
            if error_model is not None:
                error_info = (
                    f"The 'model' parameter must be a {target_module_type[1]} or list[{target_module_type[1]}] "
                    f"type, currently there is a {type(error_model)} type.")
                raise MsprobeException(
                    MsprobeException.INVALID_PARAM_ERROR, error_info)

        else:
            error_info = (f"The 'model' parameter must be a {target_module_type[1]} or list[{target_module_type[1]}] "
                          f"type, currently there is a {type(models)} type.")
            raise MsprobeException(
                MsprobeException.INVALID_PARAM_ERROR, error_info)
        return models

    def check(self):
        if not self.dump_path:
            raise Exception("Dump path is empty.")
        self.dump_path = os.path.abspath(self.dump_path)
        if not self.task:
            self.task = "statistics"
        if not self.level:
            raise Exception("level must be L0, L1 or L2")
        if not self.file_format:
            self.file_format = "npy"
        if not self.check_mode:
            self.check_mode = "all"
        if not isinstance(self.async_dump, bool):
            raise Exception("The parameters async_dump should be bool.")
        if self.task == Const.STRUCTURE and self.level_ori not in [Const.LEVEL_L0, Const.LEVEL_MIX]:
            logger.warning_on_rank_0(
                f"When the task is set to structure, the level should be one of {[Const.LEVEL_L0, Const.LEVEL_MIX]}. "
                f"If not, the default level is {Const.LEVEL_MIX}."
            )
            self.level_ori = Const.LEVEL_MIX
        if self.async_dump:
            if self.task == Const.TENSOR:
                if self.level_ori == Const.LEVEL_DEBUG:
                    self.list = []  # async_dump + debug level case ignore list
                if not self.list and self.level_ori != Const.LEVEL_DEBUG:
                    raise MsprobeException(
                        MsprobeException.INVALID_PARAM_ERROR,
                        "The parameters async_dump is true in tensor task, the parameters list cannot be empty."
                    )
            is_unsupported_mode = self.summary_mode == Const.MD5 or \
                                  isinstance(self.summary_mode, list) and Const.MD5 in self.summary_mode
            if is_unsupported_mode:
                raise MsprobeException(
                    MsprobeException.INVALID_PARAM_ERROR,
                    f"The parameters async_dump is true, the parameters summary_mode cannot be/contain md5."
                )
        return True

    def check_config_with_l2(self, is_graph_config):
        if not is_graph_config and self.task != Const.TENSOR:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"When level is set to L2, the task must be set to tensor.")
        if not is_graph_config and self.scope:
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"When level is set to L2, the scope cannot be configured.")
        if not is_graph_config and (not self.list or len(self.list) != 1):
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR,
                                   f"When level is set to L2, the list must be configured as a list with one api name.")

    def _check_statistics_config(self, task_config):
        if self.task != Const.STATISTICS:
            return
        self.tensor_list = []
        if not hasattr(task_config, "tensor_list"):
            return
        if self.level_ori == Const.LEVEL_DEBUG and task_config.tensor_list:
            logger.warning_on_rank_0("When level is set to debug, the tensor_list will be invalid.")
            return
        self.tensor_list = task_config.tensor_list



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class mainly dump model ops inputs and outputs.
"""

import os
import time

from msprobe.core.common.file_utils import check_file_or_directory_path, create_directory
from msprobe.core.common.log import logger
from msprobe.core.common.const import Const
from msprobe.infer.offline.compare.msquickcmp.atc import atc_utils
from msprobe.infer.offline.compare.msquickcmp.common import utils
from msprobe.infer.offline.compare.msquickcmp.common.convert import convert_npy_to_bin
from msprobe.infer.offline.compare.msquickcmp.common.utils import AccuracyCompareException, get_shape_to_directory_name
from msprobe.infer.offline.compare.msquickcmp.dump.args_adapter import DumpArgsAdapter
from msprobe.infer.offline.compare.msquickcmp.npu.npu_dump_data import NpuDumpData
from msprobe.infer.offline.compare.msquickcmp.npu.om_parser import OmParser


def dump_process(args: DumpArgsAdapter):
    """
    Function Description:
        main process function
    Exception Description:
        exit the program when an AccuracyCompare Exception  occurs
    """
    args.golden_path = os.path.realpath(args.golden_path)
    args.cann_path = os.path.realpath(args.cann_path)
    args.input_data = convert_npy_to_bin(args.input_data)
    check_and_dump(args)
    logger.info(Const.TOOL_ENDS_SUCCESSFULLY)


def check_and_dump(args):
    check_file_or_directory_path(args.golden_path, False)
    utils.check_device_param_valid(args.rank)
    time_dir = time.strftime("%Y%m%d%H%M%S", time.localtime())
    original_out_path = os.path.realpath(os.path.join(args.output_path, time_dir))
    args.output_path = original_out_path
    create_directory(args.output_path)
    # deal with the dymShape_range param if exists
    input_shapes = []
    if args.dym_shape_range:
        input_shapes = utils.parse_dym_shape_range(args.dym_shape_range)
    if not input_shapes:
        input_shapes.append("")
    for input_shape in input_shapes:
        dump_data(args, input_shape, original_out_path)


def dump_data(args, input_shape, original_out_path):
    if input_shape:
        args.input_shape = input_shape
        args.output_path = os.path.join(original_out_path, get_shape_to_directory_name(args.input_shape))

    _, extension = utils.get_model_name_and_extension(args.golden_path)
    if extension == ".onnx":
        from msprobe.infer.offline.compare.msquickcmp.onnx_model.onnx_dump_data import OnnxDumpData
        dumper = OnnxDumpData(args, npu_dump_npy_path=None)
        dumper.generate_inputs_data(npu_dump_data_path=None, use_aipp=False)
        dumper.generate_dump_data()
    elif extension == ".om":
        output_json_path = atc_utils.convert_model_to_json(args.cann_path, args.golden_path, args.output_path)
        om_parser = OmParser(output_json_path)
        use_aipp = True if om_parser.get_aipp_config_content() else False

        args.target_path = args.golden_path
        dumper = NpuDumpData(args, is_golden=False)
        dumper.generate_inputs_data(use_aipp=use_aipp)
        dumper.generate_dump_data()
    else:
        logger.error("Dump model files whose names end with .onnx or .om are supported, please check your model type.")
        raise AccuracyCompareException(utils.ACCURACY_COMPARISON_MODEL_TYPE_ERROR)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.pytorch.api_accuracy_checker.compare.algorithm import get_rel_err_ratio
from msprobe.core.common.const import CompareConst
from msprobe.pytorch.api_accuracy_checker.precision_standard.base_standard import BaseCompare


class ThousandthStdCompare(BaseCompare):
    """
    Thousandth standard comparison class for calculating accuracy metrics.
    
    A subclass of BaseCompare, specifically designed to compare the relative error 
    between benchmark and device outputs, focusing on errors within a thousandth (0.001) threshold.

    Attributes:
        rel_err_orign (float or array-like): The original relative error values to be compared.
        compare_column (object): An object to store and update comparison metrics.

    Methods:
        _compute_metrics(): Computes the relative error metrics, specifically the thousandth error ratio.
    """
    def __init__(self, input_data):
        super().__init__(input_data)
        self.rel_err_orign = input_data.rel_err_orign
        self.compare_column = input_data.compare_column

    def _pre_compare(self):
        pass

    def _compute_metrics(self):
        """
        Computes the relative error metrics for the comparison, specifically focusing on errors within a thousandth 
        (0.001) threshold.

        This method calculates the proportion of relative errors that are within the thousandth threshold.
        It uses the `get_rel_err_ratio` function to determine the ratio of relative errors that are less than or 
        equal to the
        specified threshold defined in `CompareConst.THOUSAND_RATIO_THRESHOLD`.

        Returns:
        dict: A dictionary containing the computed relative error metric.
            The dictionary has the following key:
            - 'rel_err_thousandth': The proportion of relative errors within the thousandth threshold.
        """
        rel_err_thousandth, _ = get_rel_err_ratio(self.rel_err_orign, CompareConst.THOUSAND_RATIO_THRESHOLD)

        return {
            'rel_err_thousandth': rel_err_thousandth
        }



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch


def npu_fast_gelu(input0):
    attr = 1.702
    const_0 = 0 - attr
    const_1 = 1
    const_2 = attr / 2

    abs_x = torch.abs(input0)
    mul_abs_x = abs_x * const_0
    exp_abs_x = torch.exp(mul_abs_x)
    div_down = exp_abs_x + const_1

    pn_x = input0 - abs_x
    mul_pn_x = pn_x * const_2
    exp_pn_x = torch.exp(mul_pn_x)
    div_up = input0 * exp_pn_x
    div_down_rec = torch.reciprocal(div_down)
    result = div_up * div_down_rec

    return result


def npu_fast_gelu_backward(grad, input_x):
    const_2 = 1.702
    const_3 = 1.0
    const_1 = 0.0 - const_2

    # e^(-1.702x)
    abs_x = torch.abs(input_x)
    mul_abs_x = abs_x * const_1
    exp_x = torch.exp(mul_abs_x)

    # 1.702xe^(-1.702x)
    add_2 = input_x * exp_x
    add_2 = add_2 * const_2

    # e^(1.702(x-|x|))
    pn_x = input_x - abs_x
    mul_pn_x = pn_x * const_2
    exp_pn_x = torch.exp(mul_pn_x)

    #  e^(-1.702x) + 1.702xe^(-1.702x) + e^(1.702(x-|x|))
    div_up = exp_x + add_2
    div_up = div_up + exp_pn_x

    # (e^(-1.702x)+1)^2
    div_down_i = exp_x + const_3
    div_down = div_down_i * div_down_i
    div_down_rec = torch.reciprocal(div_down)
    result_temp = div_up * div_down_rec
    result = grad * result_temp

    return result.cpu()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


from __future__ import annotations

from typing import Any, Dict, Optional, Type

from msprobe.core.common.log import logger
from msprobe.core.monitor_v2.module import (
    MindSporeModuleMonitorV2,
    PyTorchModuleMonitorV2,
)
from msprobe.core.monitor_v2.cc import PyTorchCCMonitorV2, MindSporeCCMonitorV2
from msprobe.core.monitor_v2.optimizer import (
    MindSporeOptimizerMonitorV2,
    PyTorchOptimizerMonitorV2,
)
from msprobe.core.monitor_v2.param import (
    MindSporeParamMonitorV2,
    PyTorchParamMonitorV2,
)
from msprobe.core.monitor_v2.weight_grad import (
    MindSporeWeightGradMonitorV2,
    PyTorchWeightGradMonitorV2,
)

MonitorCls = Type


class MonitorFactory:
    """
    Registry-based factory that organizes monitors by framework, then by monitor name.
    """

    _REGISTRY: Dict[str, Dict[str, MonitorCls]] = {
        "pytorch": {
            "module": PyTorchModuleMonitorV2,
            "weight_grad": PyTorchWeightGradMonitorV2,
            "optimizer": PyTorchOptimizerMonitorV2,
            "param": PyTorchParamMonitorV2,
            "cc": PyTorchCCMonitorV2,
        },
        "mindspore": {
            "module": MindSporeModuleMonitorV2,
            "weight_grad": MindSporeWeightGradMonitorV2,
            "optimizer": MindSporeOptimizerMonitorV2,
            "param": MindSporeParamMonitorV2,
            "cc": MindSporeCCMonitorV2,
        },
    }

    @classmethod
    def create(cls, framework: str, name: str) -> Optional[Any]:
        fwk_key = str(framework).lower()
        name_key = str(name).lower()
        fwk = cls._REGISTRY.get(fwk_key)
        if not fwk:
            logger.warning(f"[monitor_v2] Unsupported framework '{framework}'.")
            return None
        mon_cls = fwk.get(name_key)
        if mon_cls is None:
            logger.warning(f"[monitor_v2] Unsupported monitor '{name}' for framework '{framework}'.")
        return mon_cls() if mon_cls else None

    @classmethod
    def available(cls, framework: str) -> Dict[str, MonitorCls]:
        return cls._REGISTRY.get(str(framework).lower(), {}).copy()



# From PyTorch:

# Copyright (c) 2025      Huawei Technologies Co., Ltd
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

# From Caffe2:

# Copyright (c) 2016-present, Facebook Inc. All rights reserved.

# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.

# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.

# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.

# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain

# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.

# All contributions by Tri Dao:
# Copyright (c) 2024 Tri Dao.
# All rights reserved.

# All contributions by Arm:
# Copyright (c) 2021, 2023-2024 Arm Limited and/or its affiliates

# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.

# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.

# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# 3. Neither the names of Facebook, Deepmind Technologies, NYU, NEC Laboratories
#    America, IDIAP Research Institute and Huawei nor the names of its contributors
#    may be used to endorse or promote products derived from this software without
#    specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings

import mindspore as ms
from mindspore.ops.operations import _inner_ops as inner
from torch.nn.modules.module import (_global_backward_pre_hooks, _global_backward_hooks,
                                     _global_is_full_backward_hook, _global_forward_pre_hooks,
                                     _global_forward_hooks, _global_forward_hooks_always_called)
from torch.utils.hooks import RemovableHandle

from msprobe.mindspore.common.utils import is_backward_hook_output_a_view


def _call_impl(self, *args, **kwargs):
    forward_call = self.forward
    if self.__ms_class__:
        return forward_call(*args, **kwargs)

    # If we don't have any hooks, we want to skip the rest of the logic in
    # this function, and just call forward.
    if not (self._backward_hooks or self._backward_pre_hooks or self._forward_hooks or self._forward_pre_hooks
            or _global_backward_pre_hooks or _global_backward_hooks
            or _global_forward_hooks or _global_forward_pre_hooks):
        return forward_call(*args, **kwargs)

    try:
        result = None
        called_always_called_hooks = set()

        if self._backward_pre_hooks or _global_backward_pre_hooks:
            _get_backward_pre_hooks(self)

        if self._backward_hooks or _global_backward_hooks:
            _get_backward_hooks(self)

        if _global_forward_pre_hooks or self._forward_pre_hooks:
            for hook_id, hook in (
                *_global_forward_pre_hooks.items(),
                *self._forward_pre_hooks.items(),
            ):
                if hook_id in self._forward_pre_hooks_with_kwargs:
                    args_kwargs_result = hook(self, args, kwargs)  # type: ignore[misc]
                    if args_kwargs_result is not None:
                        if isinstance(args_kwargs_result, tuple) and len(args_kwargs_result) == 2:
                            args, kwargs = args_kwargs_result
                        else:
                            raise RuntimeError(
                                "forward pre-hook must return None or a tuple "
                                f"of (new_args, new_kwargs), but got {args_kwargs_result}."
                            )
                else:
                    args_result = hook(self, args)
                    if args_result is not None:
                        if not isinstance(args_result, tuple):
                            args_result = (args_result,)
                        args = args_result

        bw_hook = None
        if self._backward_hooks:
            bw_hook = inner.CellBackwardHook(self.__class__.__name__ + "(" + str(id(self)) + ")",
                                             self, self._backward_hooks)
            bw_hook.register_backward_hook()
            args = apply_backward_hook_on_tensors(bw_hook, args)

        result = forward_call(*args, **kwargs)
        if _global_forward_hooks or self._forward_hooks:
            for hook_id, hook in (
                *_global_forward_hooks.items(),
                *self._forward_hooks.items(),
            ):
                # mark that always called hook is run
                if hook_id in self._forward_hooks_always_called or hook_id in _global_forward_hooks_always_called:
                    called_always_called_hooks.add(hook_id)

                if hook_id in self._forward_hooks_with_kwargs:
                    hook_result = hook(self, args, kwargs, result)
                else:
                    hook_result = hook(self, args, result)

                if hook_result is not None:
                    result = hook_result

        if bw_hook:
            if not isinstance(result, (ms.Tensor, tuple)):
                warnings.warn("For backward hooks to be called,"
                              " module output should be a Tensor or a tuple of Tensors"
                              f" but received {type(result)}")
            result = apply_backward_hook_on_tensors(bw_hook, result)

        if self._backward_pre_hooks:
            bw_pre_hook = inner.CellBackwardHook(self.__class__.__name__ + "(" + str(id(self)) + ")",
                                                 self, self._backward_pre_hooks)
            bw_pre_hook.register_backward_pre_hook()
            result = apply_backward_hook_on_tensors(bw_pre_hook, result)

        return result
    except Exception:
        # run always called hooks if they have not already been run
        # For now only forward hooks have the always_call option but perhaps
        # this functionality should be added to full backward hooks as well.
        for hook_id, hook in _global_forward_hooks.items():
            # type: ignore[possibly-undefined]
            if hook_id in _global_forward_hooks_always_called and hook_id not in called_always_called_hooks:
                try:
                    hook_result = hook(self, args, result)  # type: ignore[possibly-undefined]
                    if hook_result is not None:
                        result = hook_result
                except Exception as e:
                    warnings.warn("global module forward hook with ``always_call=True`` raised an exception "
                                  f"that was silenced as another error was raised in forward: {str(e)}")
                    continue

        for hook_id, hook in self._forward_hooks.items():
            # type: ignore[possibly-undefined]
            if hook_id in self._forward_hooks_always_called and hook_id not in called_always_called_hooks:
                try:
                    if hook_id in self._forward_hooks_with_kwargs:
                        hook_result = hook(self, args, kwargs, result)  # type: ignore[possibly-undefined]
                    else:
                        hook_result = hook(self, args, result)  # type: ignore[possibly-undefined]
                    if hook_result is not None:
                        result = hook_result
                except Exception as e:
                    warnings.warn("module forward hook with ``always_call=True`` raised an exception "
                                  f"that was silenced as another error was raised in forward: {str(e)}")
                    continue
        # raise exception raised in try block
        raise


def register_full_backward_pre_hook(self, hook, prepend: bool = False) -> RemovableHandle:
    handle = RemovableHandle(self._backward_pre_hooks)
    self._backward_pre_hooks[handle.id] = hook
    if prepend:
        self._backward_pre_hooks.move_to_end(handle.id, last=False)  # type: ignore[attr-defined]
    return handle


def register_full_backward_hook(self, hook, prepend: bool = False) -> RemovableHandle:
    if self._is_full_backward_hook is False:
        raise RuntimeError(
            "Cannot use both regular backward hooks and full backward hooks on a "
            "single Module. Please use only one of them."
        )

    self._is_full_backward_hook = True

    handle = RemovableHandle(self._backward_hooks)
    self._backward_hooks[handle.id] = hook
    if prepend:
        self._backward_hooks.move_to_end(handle.id, last=False)  # type: ignore[attr-defined]
    return handle


def _get_backward_pre_hooks(self):
    self._backward_pre_hooks.update(_global_backward_pre_hooks)


def _get_backward_hooks(self):
    if (_global_is_full_backward_hook is True):
        self._backward_hooks.update(_global_backward_hooks)


def apply_backward_hook_on_tensors(cell_backward_hook, args):
    if is_backward_hook_output_a_view():
        hooked_args = cell_backward_hook(args)
    else:
        is_tuple = True
        if not isinstance(args, tuple):
            args = (args,)
            is_tuple = False
        hooked_args = cell_backward_hook(*args)
        if is_tuple and len(args) == 1:
            hooked_args = (hooked_args, )
    return hooked_args



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
FusionOp class. This class mainly involves the fusion op info.
"""
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


class OutputDesc:
    """
    The class for fusion op output desc
    """

    def __init__(self: any, origin_name: str, origin_output_index: int, origin_format: str,
                 origin_shape: list) -> None:
        self.origin_name = origin_name
        self.origin_output_index = origin_output_index
        self.origin_format = origin_format
        self.origin_shape = origin_shape
        self.data_type = ""

    def get_origin_name(self: any) -> str:
        """
        Get origin name
        """
        return self.origin_name

    def get_origin_shape(self: any) -> list:
        """
        Get origin shape
        """
        return self.origin_shape

    def set_data_type(self: any, data_type: str) -> None:
        """
        Set the data type of output
        """
        self.data_type = data_type


class OpAttr:
    """
    The class for op attr
    """

    def __init__(self: any, original_op_names: list, l1_fusion_no: str, is_multi_op: bool, op_sequence: int) -> None:
        self.original_op_names = original_op_names
        self.l1_fusion_no = l1_fusion_no
        self._is_multi_op = is_multi_op
        self._op_sequence = op_sequence
        self.quant_filter = False

    def is_multi_op(self: any) -> bool:
        """
        is multi op
        """
        return self._is_multi_op

    def get_op_sequence(self: any) -> int:
        """
        Get op sequence
        """
        return self._op_sequence


class Tensor:
    """
    The class for tensor info
    """

    def __init__(self: any, name: str, index: int, tensor_format: str, shape: list) -> None:
        self.name = name
        self.index = index
        self.tensor_format = tensor_format
        self.shape = shape
        self.path = ''
        self.data = None

    def set_data(self: any, data: any) -> None:
        """
        Set data
        """
        self.data = data

    def set_path(self: any, path: str) -> None:
        """
        Set path
        """
        self.path = path


class FusionOp:
    """
    The class for fusion op
    """

    def __init__(self: any, *args: any) -> None:
        op_id, op_name, input_list, op_type, output_desc, attr = args
        self.op_id = op_id
        self.op_name = op_name
        self.input_list = input_list
        self.op_type = op_type
        self.output_desc = output_desc
        self.attr = attr

    def get_origin_tensor(self: any, index: int) -> Tensor:
        """
        Get the origin tensor
        :param index: the index
        :return: the Tensor
        """
        if len(self.output_desc) == 0:
            message = 'The output description of the fusion operator "%s" is empty.' % self.op_name
            log.print_warn_log(message)
            raise CompareError(CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR, message)
        real_index = index
        if real_index >= len(self.output_desc):
            real_index = 0
        origin_name = self.output_desc[real_index].origin_name
        if origin_name == "":
            message = 'The origin name of the %s:%d is empty.' % (self.op_name, index)
            log.print_warn_log(message)
            raise CompareError(CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR, message)
        origin_output_index = self.output_desc[real_index].origin_output_index
        if origin_output_index is None:
            origin_output_index = index
        return Tensor(origin_name, origin_output_index, self.output_desc[real_index].origin_format,
                      self.output_desc[real_index].origin_shape)

    def get_input_tensor(self: any, index: int) -> Tensor:
        """
        Get the input tensor
        :param index: the input index
        :return: the Tensor
        """
        if index >= len(self.input_list):
            message = 'The input index (%d) is out of range (%d) for the operator list.' % (index, len(self.input_list))
            log.print_error_log(message)
            raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)

        input_op_info = self.input_list[index].split(':')
        if len(input_op_info) != ConstManager.INPUT_INFO_COUNT:
            message = 'The input information (%s) is invalid. It only supports op_name:index.' % self.input_list[index]
            log.print_error_log(message)
            raise CompareError(CompareError.MSACCUCMP_INVALID_JSON_FILE_ERROR, message)

        input_op_name = input_op_info[ConstManager.INPUT_OP_NAME_INDEX].strip()
        if not input_op_name:
            message = 'The input information (%s) is invalid. The input operator name is empty.' \
                      % self.input_list[index]
            log.print_error_log(message)
            raise CompareError(CompareError.MSACCUCMP_INVALID_JSON_FILE_ERROR, message)

        input_op_index = input_op_info[ConstManager.INPUT_INDEX_INDEX].strip()
        if not RegManager.match_pattern(RegManager.NUMBER_PATTERN, input_op_index):
            message = 'The input information (%s) is invalid. The input index only supports numbers.' \
                      % self.input_list[index]
            log.print_error_log(message)
            raise CompareError(CompareError.MSACCUCMP_INVALID_JSON_FILE_ERROR, message)

        return Tensor(input_op_name, int(input_op_index), '', [])

    def is_inner_node(self: any) -> bool:
        """
        Check the fusion op is inner node.
        :return True, if the node is inner node
        """
        inner_node = False
        if self.attr.is_multi_op():
            inner_node = True
            for output_desc in self.output_desc:
                if output_desc.origin_name != '':
                    inner_node = False
                    break
        return inner_node

    def get_real_op_type(self: any) -> str:
        """
        if op_type is not Left or Right, there is real op type
        """
        if self.op_type not in [ConstManager.LEFT_TYPE, ConstManager.RIGHT_TYPE]:
            op_type = self.op_type
        else:
            op_type = ConstManager.NAN
        return op_type
